#imports
import re

class DataFileHandlingConstants :
    @property
    def CHUNK_ALREADY_WRITTEN_CODE(self) :
        return 10 # code indicating that a particular chunk has already been written
    @property
    def FILE_HASH_MISMATCH_CODE(self) :
        return -1 # code indicating that a file's hashes didn't match
    @property
    def FILE_SUCCESSFULLY_RECONSTRUCTED_CODE(self) :
        return 3  # code indicating that a file was successfully fully reconstructed
    @property
    def FILE_IN_PROGRESS(self) :
        return 2  # code indicating that a file is in the process of being reconstructed
    
DATA_FILE_HANDLING_CONST=DataFileHandlingConstants()

class RunOptionConstants :
    @property
    def DEFAULT_CONFIG_FILE(self) :
        return 'test' # name of the config file that will be used by default
    @property
    def DEFAULT_TOPIC_NAME(self) :
        return 'test' # name of the topic to produce to by default
    @property
    def PRODUCTION_CONFIG_FILE(self) :
        return 'prod' # name of the config file used in "real" production
    @property
    def N_DEFAULT_UPLOAD_THREADS(self) :
        return 2      # default number of threads to use when uploading a file
    @property
    def N_DEFAULT_DOWNLOAD_THREADS(self) :
        return 2      # default number of threads to use when downloading chunks of a file
    @property
    def DEFAULT_UPLOAD_REGEX(self) :
        #matches everything except something starting with a '.' or ending in '.log'
        return re.compile(r'^((?!(\.|.*.log))).*$') 
    @property
    def DEFAULT_CHUNK_SIZE(self) :
        return 16384  # default size in bytes of each file upload chunk
    @property
    def DEFAULT_MAX_UPLOAD_QUEUE_SIZE(self) :
        return 3000   # default maximum number of items allowed in the upload Queue at once

RUN_OPT_CONST = RunOptionConstants()



#imports
from abc import ABC, abstractmethod
from .controlled_process import ControlledProcess

class ControlledProcessSingleThread(ControlledProcess,ABC) :
    """
    A class for running a process in a single thread in a loop until it's explicitly shut down
    """

    def run(self) :
        """
        Start the process and call run_iteration until the process is shut down
        """
        super().run()
        while self.alive :
            self._run_iteration()
            self._print_still_alive()
            self._check_control_command_queue()

    @abstractmethod
    def _run_iteration(self) :
        """
        The function that is run in an infinite loop while the process is alive
        Not implemented in the base class.
        """
        pass



#imports
from ..utilities.misc import populated_kwargs
from ..utilities.logging import LogOwner

class DataFileDirectory(LogOwner) :
    """
    Base class representing any directory holding data files
    """

    @property
    def dirpath(self) :
        return self.__dirpath
    @property
    def data_files_by_path(self) :
        return self.__data_files_by_path

    def __init__(self,dirpath,*args,**kwargs) :
        """
        dirpath = path to the directory 
        """
        self.__dirpath = dirpath.resolve()
        self.__data_files_by_path = {}
        kwargs = populated_kwargs(kwargs,{'logger_file':self.__dirpath})
        super().__init__(*args,**kwargs)



#imports
from abc import ABC, abstractmethod

class HasArgumentParser(ABC) :
    """
    Base class for objects that have argument parsers
    """

    @classmethod
    @abstractmethod
    def get_argument_parser(cls,*args,**kwargs) :
        """
        Return the argument parser that objects of this class use
        Not implemented in the base class
        """
        pass



# @author: Amir H. Sharifzadeh, https://pages.jh.edu/asharif9/
# @project: OpenMSI
# @date: 04/12/2022
# @owner: The Institute of Data Intensive Engineering and Science, https://idies.jhu.edu/
# Johns Hopkins University http://www.jhu.edu
# import logging

from botocore.exceptions import ClientError
from .s3_service import S3Service

class S3DataTransfer(S3Service) :

    def __init__(self, s3_config, *args, **kwargs):
        super().__init__(s3_config,*args,**kwargs)

    def transfer_object_stream(self, topic_name,datafile):
        file_name = str(datafile.filename)
        sub_dir=datafile.subdir_str
        s3_full_path = topic_name 
        if sub_dir is not None :
            s3_full_path+= '/' + sub_dir 
        s3_full_path+= '/' + file_name
        try:
            self.s3_client.put_object(Body=datafile.bytestring, Bucket=self.bucket_name,
                                      Key=s3_full_path,
                                      # GrantRead=self.grant_read
                                      )
            msg = f'{file_name} successfully transferred into {self.bucket_name}/{s3_full_path}'
            self.logger.info(msg)
        except ClientError as e:
            self.logger.error(f'{e.response}: failed to transfer {file_name} into {self.bucket_name}/{s3_full_path}')



#imports
from ..utilities.logging import LogOwner
from .openmsistream_producer import OpenMSIStreamProducer

class ProducerGroup(LogOwner) :
    """
    Class for working with a group of producers sharing a single KafkaCrypto instance
    """

    def __init__(self,config_path,*args,**kwargs) :
        """
        config_path = path to the config file that should be used to define the producer group
        """
        super().__init__(*args,**kwargs)
        self.__p_args, self.__p_kwargs = OpenMSIStreamProducer.get_producer_args_kwargs(config_path,logger=self.logger)

    def get_new_producer(self) :
        """
        Return a new Producer

        Call this function from a child thread to get thread-independent Producers

        This function just creates the Producer; closing it etc. must be handled by whatever calls this function.
        """
        producer = OpenMSIStreamProducer(*self.__p_args,**self.__p_kwargs)
        return producer

    def close(self) :
        try :
            self.__p_kwargs['kafkacrypto'].close()
        except :
            pass
        finally :
            self.__p_kwargs['kafkacrypto'] = None


#imports
from abc import ABC, abstractmethod
from threading import Lock
from kafkacrypto.message import KafkaCryptoMessage
from ..kafka_wrapper.controlled_message_processor import ControlledMessageProcessor
from .config import DATA_FILE_HANDLING_CONST
from .data_file_chunk import DataFileChunk
from .download_data_file import DownloadDataFile

class DataFileChunkProcessor(ControlledMessageProcessor,ABC) :
    """
    Use a ControlledProcessMultiThreaded interface to process messages that are expected to be DataFileChunks
    """

    #################### PROPERTIES ####################

    @property
    def progress_msg(self) :
        progress_msg = 'The following files have been recognized so far:\n'
        for datafile in self.files_in_progress_by_path.values() :
            progress_msg+=f'\t{datafile.full_filepath} (in progress)\n'
        for fp in self.completely_processed_filepaths :
            progress_msg+=f'\t{fp} (completed)\n'
        return progress_msg

    @property
    def other_datafile_kwargs(self) :
        return {} #Overload this in child classes to define additional keyword arguments 
                  #that should go to the specific datafile constructor
    @property
    def completely_processed_filepaths(self) :
        return self.__completely_processed_filepaths

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,datafile_type,**kwargs) :
        """
        datafile_type = the type of datafile that the consumed messages will be used to create 
            (must be a subclass of DownloadDataFile)
        """    
        super().__init__(*args,**kwargs)
        self.datafile_type = datafile_type
        if not issubclass(self.datafile_type,DownloadDataFile) :
            errmsg = 'ERROR: DataFileChunkProcessor requires a datafile_type that is a subclass of '
            errmsg+= f'DownloadDataFile but {self.datafile_type} was given!'
            raise ValueError(errmsg)
        self.files_in_progress_by_path = {}
        self.locks_by_fp = {}
        self.__completely_processed_filepaths = []

    #################### PRIVATE HELPER FUNCTIONS ####################

    @abstractmethod
    def _process_message(self, lock, msg, rootdir_to_set, logger):
        """
        Make sure message values are of the expected DataFileChunk type with no root directory set, and then 
        add the chunk to the data file object. If the file is in progress this function returns True.
        Otherwise the code from DownloadDataFile.add_chunk will be returned.

        If instead the message was encrypted and could not be successfully decrypted, this will return 
        the raw Message object with KafkaCryptoMessages as its key and/or value

        lock = the Thread Lock object to use when processing the file this message comes from
        msg = the actual message object from a call to consumer.poll
        rootdir_to_set = root directory for the new DataFileChunk
        logger = the logger object to use
        
        Child classes should call super()._process_message() before doing anything else with
        the message to perform these checks
        """
        # If the message has KafkaCryptoMessages as its key and/or value, then decryption failed. 
        # Return the message object instead of a code.
        if ( hasattr(msg,'key') and hasattr(msg,'value') and 
             (isinstance(msg.key,KafkaCryptoMessage) or isinstance(msg.value,KafkaCryptoMessage)) ) :
            return msg
        #get the DataFileChunk from the message value
        try :
            dfc = msg.value() #from a regular Kafka Consumer
        except :
            dfc = msg.value #from KafkaCrypto
        #make sure the chunk is of the right type
        if not isinstance(dfc,DataFileChunk) :
            errmsg = f'ERROR: expected DataFileChunk messages but received a message of type {type(dfc)}!'
            self.logger.error(errmsg,ValueError)
        #make sure the chunk doesn't already have a rootdir set
        if dfc.rootdir is not None :
            errmsg = f'ERROR: message with key {dfc.message_key} has rootdir={dfc.rootdir} '
            errmsg+= '(should be None as it was just consumed)! Will ignore this message and continue.'
            self.logger.error(errmsg)
        #set the chunk's root directory
        dfc.rootdir = rootdir_to_set
        #add the chunk's data to the file that's being reconstructed
        with lock :
            if dfc.filepath not in self.files_in_progress_by_path.keys() :
                self.files_in_progress_by_path[dfc.filepath] = self.datafile_type(dfc.filepath,
                                                                                  logger=logger,
                                                                                  **self.other_datafile_kwargs)
                self.locks_by_fp[dfc.filepath] = Lock()
        retval = self.files_in_progress_by_path[dfc.filepath].add_chunk(dfc,self.locks_by_fp[dfc.filepath])
        #If the file is just in progress, return True
        if retval in (DATA_FILE_HANDLING_CONST.FILE_IN_PROGRESS,
                            DATA_FILE_HANDLING_CONST.CHUNK_ALREADY_WRITTEN_CODE) :
            return True
        #otherwise just return the code from add_chunk
        return retval



#imports
import time, traceback
from abc import ABC, abstractmethod
from ..running.controlled_process_multi_threaded import ControlledProcessMultiThreaded
from .consumer_group import ConsumerGroup

class ControlledMessageProcessor(ControlledProcessMultiThreaded,ConsumerGroup,ABC) :
    """
    Combine a ControlledProcessMultiThreaded and a ConsumerGroup to create a 
    single interface for reading and processing individual messages
    """

    CONSUMER_POLL_TIMEOUT = 0.050
    NO_MESSAGE_WAIT = 0.005 #how long to wait if consumer.get_next_message_value returns None

    def __init__(self,*args,**kwargs) :
        """
        Hang onto the number of messages read and processed
        """
        self.n_msgs_read = 0
        self.n_msgs_processed = 0
        super().__init__(*args,**kwargs)

    def _run_worker(self,lock):
        """
        Handle startup and shutdown of a thread-independent Consumer and 
        serve individual messages to the _process_message function
        """
        #create the Consumer for this thread
        if self.alive :
            consumer = self.get_new_subscribed_consumer()
        if ('enable.auto.commit' not in consumer.configs.keys()) or (consumer.configs['enable.auto.commit'] is True) :
            warnmsg = 'WARNING: enable.auto.commit has not been set to False for a Consumer that will manually commit '
            warnmsg+= 'offsets. Missed or duplicate messages could result. You can set "enable.auto.commit"=False in '
            warnmsg+= 'the "consumer" section of the config file to re-enable manual offset commits (recommended).'
            self.logger.warning(warnmsg)
        #start the loop for while the controlled process is alive
        last_message = None
        while self.alive :
            #consume a message from the topic
            msg = consumer.get_next_message(ControlledMessageProcessor.CONSUMER_POLL_TIMEOUT)
            if msg is None :
                time.sleep(ControlledMessageProcessor.NO_MESSAGE_WAIT) #wait just a bit to not over-tax things
                continue
            with lock :
                self.n_msgs_read+=1
            last_message = msg
            #send the message to the _process_message function
            retval = self._process_message(lock,msg)
            #count and (asynchronously) commit the message as processed
            if retval :
                tps = consumer.commit(msg)
                if tps is not None :
                    for tp in tps :
                        if tp.error is not None :
                            warnmsg = 'WARNING: failed to synchronously commit offset of last message received on '
                            warnmsg+= f'"{tp.topic}" partition {tp.partition}. Duplicate messages may result when this '
                            warnmsg+= f'Consumer is restarted. Error reason: {tp.error.str()}'
                            self.logger.warning(warnmsg)
                with lock :
                    self.n_msgs_processed+=1
        #commit the offset of the last message received (block until done)
        if last_message is not None :
            try :
                tps = consumer.commit(last_message,asynchronous=False)
                if tps is not None :
                    for tp in tps :
                        if tp.error is not None :
                            warnmsg = 'WARNING: failed to synchronously commit offset of last message received on '
                            warnmsg+= f'"{tp.topic}" partition {tp.partition}. Duplicate messages may result when this '
                            warnmsg+= f'Consumer is restarted. Error reason: {tp.error.str()}'
                            self.logger.warning(warnmsg)
            except Exception as e :
                warnmsg = 'WARNING: failed to synchronously commit offset of last message received. '
                warnmsg+= 'Duplicate messages may be read the next time this Consumer is started.'
                try :
                    raise e
                except Exception :
                    warnmsg+= f' Error traceback: {traceback.format_exc()}'
                self.logger.warning(warnmsg)
        #shut down the Consumer that was created once the process isn't alive anymore
        consumer.close()

    def _on_shutdown(self):
        super()._on_shutdown()
        self.close()

    @abstractmethod
    def _process_message(self,lock,msg,*args,**kwargs) :
        """
        Process a single message read from the thread-independent Consumer
        Returns true if processing was successful, and False otherwise
        
        lock = lock across all created child threads (use to enforce thread safety during processing)
        msg  = a single message that was consumed and should be processed by this function

        Not implemented in the base class 
        """
        pass



from ..running.runnable import Runnable
from ..data_file_io.config import RUN_OPT_CONST
from ..data_file_io.data_file_stream_processor import DataFileStreamProcessor
from .config_file_parser import S3ConfigFileParser
from .s3_data_transfer import S3DataTransfer

class S3TransferStreamProcessor(DataFileStreamProcessor, Runnable) :

    def __init__(self, bucket_name, config_path, topic_name, **otherkwargs) :
        super().__init__(config_path, topic_name, **otherkwargs)
        parser = S3ConfigFileParser(config_path,logger=self.logger)
        self.__s3_config = parser.s3_configs
        self.__s3_config['bucket_name'] = bucket_name
        self.bucket_name = bucket_name
        self.s3d = S3DataTransfer(self.__s3_config,logger=self.logger)

    def make_stream(self):
        return self.process_files_as_read()

    def stream_processor_run(self) :
        self.make_stream()

    def close_session_client(self) :
        self.s3d.close_session()

    def _process_downloaded_data_file(self, datafile, lock):
        self.s3d.transfer_object_stream(self.topic_name, datafile)
        if self.s3d.compare_consumer_datafile_with_s3_object_stream(self.topic_name, self.bucket_name, datafile):
            file_name = str(datafile.filename)
            sub_dir = datafile.subdir_str
            object_key = self.topic_name 
            if sub_dir is not None :
                object_key+= '/' + sub_dir 
            object_key+= '/' + file_name
            self.logger.info(object_key + ' matched with consumer datafile')
            # self.s3d.delete_object_from_bucket(self.bucket_name, object_key)
        return None

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [*superargs, 'bucket_name',
                'update_seconds', 'logger_file', 'config', 'topic_name', 'consumer_group_ID']
        kwargs = {**superkwargs,
                  'n_threads': RUN_OPT_CONST.N_DEFAULT_DOWNLOAD_THREADS, 
                  }
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        # make the argument parser
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        s3_stream_proc = cls(args.bucket_name,
                             args.config, args.topic_name,
                             n_threads=args.n_threads,
                             update_secs=args.update_seconds,
                             consumer_group_ID=args.consumer_group_ID,
                             logger_file=args.logger_file)
        # cls.bucket_name = args.bucket_name
        msg = f'Listening to the {args.topic_name} topic for files to add to the {args.bucket_name} bucket....'
        s3_stream_proc.logger.info(msg)
        n_read,n_processed,complete_filenames = s3_stream_proc.make_stream()
        s3_stream_proc.close()
        msg = f'{n_read} total messages were consumed'
        if len(complete_filenames)>0 :
            msg+=f', {n_processed} messages were successfully processed,'
            msg+=f' and the following {len(complete_filenames)} file'
            msg+=' was' if len(complete_filenames)==1 else 's were'
            msg+=f' successfully transferred to the {args.bucket_name} bucket'
            for fn in complete_filenames :
                msg+=f'\n\t{fn}'
        else :
            msg+=f' and {n_processed} messages were successfully processed'
        s3_stream_proc.logger.info(msg)

    def _on_check(self) :
        msg = f'{self.n_msgs_read} messages read, {self.n_msgs_processed} messages processed, '
        msg+= f'{len(self.completely_processed_filepaths)} files transferred so far'
        self.logger.debug(msg)
        if len(self.files_in_progress_by_path)>0 or len(self.completely_processed_filepaths)>0 :
            self.logger.debug(self.progress_msg)

#################### MAIN METHOD TO RUN FROM COMMAND LINE ####################

def main(args=None):
    S3TransferStreamProcessor.run_from_command_line(args)

if __name__ == '__main__':
    main()



#imports
import pathlib, datetime, time
from threading import Lock
from queue import Queue
from ..utilities.misc import populated_kwargs
from ..utilities.exception_tracking_thread import ExceptionTrackingThread
from ..running.runnable import Runnable
from ..running.controlled_process_single_thread import ControlledProcessSingleThread
from ..kafka_wrapper.producer_group import ProducerGroup
from .config import RUN_OPT_CONST
from .producer_file_registry import ProducerFileRegistry
from .data_file_directory import DataFileDirectory
from .upload_data_file import UploadDataFile

class DataFileUploadDirectory(DataFileDirectory,ControlledProcessSingleThread,ProducerGroup,Runnable) :
    """
    Class representing a directory being watched for new files to be added so they can be uploaded
    """

    #################### CONSTANTS AND PROPERTIES ####################

    MIN_WAIT_TIME = 0.005 # starting point for how long to wait between pinging the directory looking for new files
    MAX_WAIT_TIME = 60 # never wait more than a minute to check again for new files
    LOG_SUBDIR_NAME = 'LOGS'

    @property
    def other_datafile_kwargs(self) :
        return {} # Overload this in subclasses to send extra keyword arguments to the individual datafile constructors
    @property
    def progress_msg(self) :
        self.__find_new_files()
        progress_msg = 'The following files have been recognized so far:\n'
        for datafile in self.data_files_by_path.values() :
            if not datafile.to_upload :
                continue
            progress_msg+=f'\t{datafile.upload_status_msg}\n'
        return progress_msg
    @property
    def have_file_to_upload(self) :
        for datafile in self.data_files_by_path.values() :
            if datafile.upload_in_progress or datafile.waiting_to_upload :
                return True
        return False
    @property
    def partially_done_file_paths(self) :
        return [fp for fp,df in self.data_files_by_path.items() if df.upload_in_progress]
    @property
    def n_partially_done_files(self) :
        return len(self.partially_done_file_paths)

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,upload_regex=RUN_OPT_CONST.DEFAULT_UPLOAD_REGEX,datafile_type=UploadDataFile,**kwargs) :
        """
        upload_regex = only files matching this regular expression will be uploaded
        datafile_type = the type of data file that recognized files should be uploaded as 
            (must be a subclass of UploadDataFile)
        """
        #create a subdirectory for the logs
        self.__logs_subdir = args[0]/self.LOG_SUBDIR_NAME
        if not self.__logs_subdir.is_dir() :
            self.__logs_subdir.mkdir(parents=True)
        #put the log file in the subdirectory
        kwargs = populated_kwargs(kwargs,{'logger_file':self.__logs_subdir})
        super().__init__(*args,**kwargs)
        if not issubclass(datafile_type,UploadDataFile) :
            errmsg = 'ERROR: DataFileUploadDirectory requires a datafile_type that is a subclass of '
            errmsg+= f'UploadDataFile but {datafile_type} was given!'
            self.logger.error(errmsg,ValueError)
        self.__upload_regex = upload_regex
        self.__datafile_type = datafile_type
        self.__wait_time = self.MIN_WAIT_TIME
        self.__lock = Lock()
        
    def upload_files_as_added(self,topic_name,
                              n_threads=RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS,
                              chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE,
                              max_queue_size=RUN_OPT_CONST.DEFAULT_MAX_UPLOAD_QUEUE_SIZE,
                              upload_existing=False) :
        """
        Listen for new files to be added to the directory. Chunk and produce newly added files as messages to the topic.

        topic_name       = name of the topic to produce messages to
        n_threads        = the number of threads to use to produce from the shared queue
        chunk_size       = the size of each file chunk in bytes
        max_queue_size   = maximum number of items allowed to be placed in the upload queue at once
        upload_existing  = set to True if any files that already exist in the directory should be uploaded
                           i.e., if False (the default) then only files added to the directory after startup
                           will be enqueued to the producer
        """
        #set the topic name and chunk size
        self.__topic_name = topic_name
        self.__chunk_size = chunk_size
        #start up a file registry in the watched directory
        self.__file_registry = ProducerFileRegistry(dirpath=self.__logs_subdir,topic_name=topic_name,logger=self.logger)
        #if we're only going to upload new files, exclude what's already in the directory
        if not upload_existing :
            self.__find_new_files(to_upload=False)
        #add or modify datafiles to upload based on the contents of the file registry
        self.__startup_from_file_registry()
        #start the upload queue and thread
        msg = 'Will upload '
        if not upload_existing :
            msg+='new files added to '
        else :
            msg+='files in '
        msg+=f'{self.dirpath} to the {topic_name} topic using {n_threads} threads'
        self.logger.info(msg)
        self.__upload_queue = Queue(max_queue_size)
        #start the producers and upload threads
        self.__producers = []
        self.__upload_threads = []
        for _ in range(n_threads) :
            self.__producers.append(self.get_new_producer())
            t = ExceptionTrackingThread(target=self.__producers[-1].produce_from_queue,
                                        args=(self.__upload_queue,self.__topic_name),
                                        kwargs={'callback': self.producer_callback},
                    )
            t.start()
            self.__upload_threads.append(t)
        #loop until the user inputs a command to stop
        self.run()
        #return a list of filepaths that have been uploaded
        return [fp for fp,datafile in self.data_files_by_path.items() if datafile.fully_produced]

    def filepath_should_be_uploaded(self,filepath) :
        """
        Filter filepaths from a glob and return a boolean that's True if they should be uploaded
        """
        if not isinstance(filepath,pathlib.PurePath) :
            self.logger.error(f'ERROR: {filepath} passed to filepath_should_be_uploaded is not a Path!',TypeError)
        if not filepath.is_file() :
            return False
        if filepath.parent == self.__logs_subdir :
            return False
        if self.__upload_regex.match(filepath.name) :
            return True
        return False

    def producer_callback(self,err,msg,filename,filepath,n_total_chunks,chunk_i) :
        """
        This function is called for every message that is acknowledged by the broker, and it uses the file registries
        to keep the information about what has and hasn't been uploaded current to what's on the broker

        err = The KafkaError object for the message
        msg = The Message object
        filename = the name of the file the message is coming from
        filepath = the full path to the file the message is coming from
        n_total_chunks = the total number of chunks in the message
        chunk_i = the index of the file chunk in the message
        """
        # If any error occured, re-enqueue the message's file chunk 
        if err is None and msg.error() is not None :
            err = msg.error()
        if err is not None : 
            if err.fatal() :
                msg =f'WARNING: fatally failed to deliver message for chunk {chunk_i} of {filepath}. '
                msg+=f'This message will be re-enqeued. Error reason: {err.str()}'
            elif not err.retriable() :
                msg =f'WARNING: Failed to deliver message for chunk {chunk_i} of {filepath} and cannot retry. '
                msg+= f'This message will be re-enqueued. Error reason: {err.str()}'
            self.logger.warning(msg)
            self.__add_chunks_for_filepath(filepath,[chunk_i])
        # Otherwise, register the chunk as successfully sent to the broker
        else :
            with self.__lock :
                fully_produced = self.__file_registry.register_chunk(filename,filepath,n_total_chunks,chunk_i)
                #If the file has now been fully produced to the topic, set the variable for the file and log a line
                if fully_produced :
                    self.data_files_by_path[filepath].fully_produced = True
                    msg = f'{filepath.relative_to(self.dirpath)} has been fully produced to the '
                    msg+= f'"{self.__topic_name}" topic as '
                    if n_total_chunks==1 :
                        msg+=f'{n_total_chunks} message'
                    else :
                        msg+=f'a set of {n_total_chunks} messages'
                    self.logger.info(msg)

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _run_iteration(self) :
        #check for new files in the directory if we haven't already found some to run
        if not self.have_file_to_upload :
            # wait here with some slight backoff so that the watched directory isn't just constantly pinged
            time.sleep(self.__wait_time) 
            self.__find_new_files()
            n_new_callbacks = 0
            for producer in self.__producers :
                new_callbacks = producer.poll(0)
                if new_callbacks is not None :
                    n_new_callbacks+=new_callbacks
                if n_new_callbacks>0 :
                    break
            if (not self.have_file_to_upload) and (n_new_callbacks==0) :
                if self.__wait_time<self.MAX_WAIT_TIME :
                    self.__wait_time*=1.1
            else :
                self.__wait_time = self.MIN_WAIT_TIME
            return
        #find the first file that's running and add some of its chunks to the upload queue 
        for datafile in self.data_files_by_path.values() :
            if datafile.upload_in_progress or datafile.waiting_to_upload :
                datafile.enqueue_chunks_for_upload(self.__upload_queue,
                                                    n_threads=len(self.__upload_threads),
                                                    chunk_size=self.__chunk_size)
        #check for new files again 
        self.__find_new_files()
        #restart any crashed threads
        self.__restart_crashed_threads()

    def _on_check(self) :
        #poll the producers
        for producer in self.__producers :
            producer.poll(0)
        #log progress so far
        self.logger.debug(self.progress_msg)
        #reset the wait time
        self.__wait_time = self.MIN_WAIT_TIME

    def _on_shutdown(self) :
        self.logger.info('Will quit after all currently enqueued files are done being transferred.')
        self.logger.info(self.progress_msg)
        #add the remainder of any files currently in progress
        if self.n_partially_done_files>0 :
            msg='Will finish queueing the remainder of the following files before flushing the producer and quitting:\n'
            for pdfp in self.partially_done_file_paths :
                msg+=f'\t{pdfp}\n'
            self.logger.info(msg)
        while self.n_partially_done_files>0 :
            for datafile in self.data_files_by_path.values() :
                if datafile.upload_in_progress :
                    datafile.enqueue_chunks_for_upload(self.__upload_queue,
                                                        n_threads=len(self.__upload_threads),
                                                        chunk_size=self.__chunk_size)
                    break
        #stop the uploading threads by adding "None"s to their queues and joining them
        for ut in self.__upload_threads :
            self.__upload_queue.put(None)
        for ut in self.__upload_threads :
            ut.join()
        self.logger.info('Waiting for all enqueued messages to be delivered (this may take a moment)....')
        for producer in self.__producers :
            producer.flush(timeout=-1) #don't move on until all enqueued messages have been sent/received
            producer.close()
        self.close()

    def __find_new_files(self,to_upload=True) :
        """
        Search the directory for any unrecognized files and add them to _data_files_by_path

        to_upload = if False, new files found will NOT be marked for uploading 
                    (default is new files are expected to be uploaded)
        """
        #This is in a try/except in case a file is moved or a subdirectory is renamed while this method is running
        #it'll just return and try again if so
        try :
            for filepath in self.dirpath.rglob('*') :
                filepath = filepath.resolve()
                if (filepath not in self.data_files_by_path.keys()) and self.filepath_should_be_uploaded(filepath) :
                    #wait until the file is actually available
                    file_ready = False
                    while not file_ready :
                        try :
                            fp = open(filepath)
                            fp.close()
                            file_ready = True
                        except PermissionError :
                            time.sleep(5.0)
                    self.data_files_by_path[filepath]=self.__datafile_type(filepath,
                                                                          to_upload=to_upload,
                                                                          rootdir=self.dirpath,
                                                                          logger=self.logger,
                                                                          **self.other_datafile_kwargs)
        except FileNotFoundError :
            return

    def __startup_from_file_registry(self) :
        """
        Look through the file registry to find and prepare to enqueue any files 
        that were in progress the last time the code was shut down and ignore 
        any files that have previously been fuly uploaded
        """
        #make sure any files listed as "in progress" have their remaining chunks enqueued
        for filepath,chunks in self.__file_registry.get_incomplete_filepaths_and_chunks() :
            self.logger.info(f'Found {filepath} in progress from a previous run. Will re-enqueue {len(chunks)} chunks.')
            self.__add_chunks_for_filepath(filepath,chunks)
        #make sure any files listed as "completed" will not be uploaded again
        for filepath in self.__file_registry.get_completed_filepaths() :
            if filepath in self.data_files_by_path.keys() :
                if self.data_files_by_path[filepath].to_upload :
                    msg = f'Found {filepath} listed as fully uploaded during a previous run. Will not produce it again.'
                    self.logger.info(msg)
                self.data_files_by_path[filepath].to_upload=False
            elif filepath.is_file() :
                msg = f'Found {filepath} listed as fully uploaded during a previous run. Will not produce it again.'
                self.logger.info(msg)
                self.data_files_by_path[filepath]=self.__datafile_type(filepath,
                                                                       to_upload=False,
                                                                       rootdir=self.dirpath,
                                                                       logger=self.logger,
                                                                       **self.other_datafile_kwargs)    

    def __add_chunks_for_filepath(self,filepath,chunks) :
        """
        Add the given chunks to the list of chunks to upload for a given file
        Creates the file if it doesn't already exist in the dictionary of recognized files
        """
        if filepath in self.data_files_by_path.keys() :
            self.data_files_by_path[filepath].to_upload=True
        else :
            self.data_files_by_path[filepath]=self.__datafile_type(filepath,
                                                                    to_upload=True,
                                                                    rootdir=self.dirpath,
                                                                    logger=self.logger,
                                                                    **self.other_datafile_kwargs)
        self.data_files_by_path[filepath].add_chunks_to_upload(chunks,chunk_size=self.__chunk_size)

    def __restart_crashed_threads(self) :
        """
        Iterate over the upload threads and restart any that crashed, logging the exception they threw
        """
        for ti,upload_thread in enumerate(self.__upload_threads) :
            if upload_thread.caught_exception is not None :
                #log the error
                warnmsg = 'WARNING: an upload thread raised an Exception, which will be logged as an error below '
                warnmsg = 'but not reraised. The Producer and upload thread that raised the error will be restarted.'
                self.logger.warning(warnmsg)
                self.logger.log_exception_as_error(upload_thread.caught_exception,reraise=False)
                #try to join the thread 
                try :
                    upload_thread.join()
                except :
                    pass
                finally :
                    self.__upload_threads[ti] = None
                    self.__producers[ti] = None
                #recreate the producer and restart the thread
                self.__producers[ti] = self.get_new_producer()
                t = ExceptionTrackingThread(target=self.__producers[ti].produce_from_queue,
                                            args=(self.__upload_queue,self.__topic_name),
                                            kwargs={'callback': self.producer_callback},
                            )
                t.start()
                self.__upload_threads[ti] = t

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls) :
        superargs,superkwargs = super().get_command_line_arguments()
        args = [*superargs,'upload_dir','config','topic_name','upload_regex',
                'chunk_size','queue_max_size','update_seconds','upload_existing']
        kwargs = {**superkwargs,'n_threads':RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls,args=None) :
        """
        Function to run the upload directory right from the command line
        """
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        #make the DataFileDirectory for the specified directory
        upload_file_directory = cls(args.upload_dir,args.config,
                                    upload_regex=args.upload_regex,update_secs=args.update_seconds)
        #listen for new files in the directory and run uploads as they come in until the process is shut down
        run_start = datetime.datetime.now()
        if not args.upload_existing :
            upload_file_directory.logger.info(f'Listening for files to be added to {args.upload_dir}...')
        else :
            upload_file_directory.logger.info(f'Uploading files in/added to {args.upload_dir}...')
        uploaded_filepaths = upload_file_directory.upload_files_as_added(args.topic_name,
                                                                         n_threads=args.n_threads,
                                                                         chunk_size=args.chunk_size,
                                                                         max_queue_size=args.queue_max_size,
                                                                         upload_existing=args.upload_existing)
        run_stop = datetime.datetime.now()
        upload_file_directory.logger.info(f'Done listening to {args.upload_dir} for files to upload')
        final_msg = f'The following {len(uploaded_filepaths)} file'
        if len(uploaded_filepaths)==1 :
            final_msg+=' was'
        else :
            final_msg+='s were'
        final_msg+=f' uploaded between {run_start} and {run_stop}:\n'
        for fp in uploaded_filepaths :
            final_msg+=f'\t{fp.relative_to(args.upload_dir)}\n'
        upload_file_directory.logger.info(final_msg)

#################### MAIN METHOD TO RUN FROM COMMAND LINE ####################

def main(args=None) :
    DataFileUploadDirectory.run_from_command_line(args)

if __name__=='__main__' :
    main()



#imports
import sys, pathlib, textwrap
from abc import abstractmethod
from subprocess import CalledProcessError
from ..utilities.logging import LogOwner
from ..utilities.config_file_parser import ConfigFileParser
from ..running.argument_parsing import OpenMSIStreamArgumentParser
from ..running.has_argument_parser import HasArgumentParser
from .config import SERVICE_CONST
from .utilities import set_env_vars, remove_env_var

class ServiceManagerBase(LogOwner,HasArgumentParser) :
    """
    Base class for working with Services in general
    """

    #################### CLASS METHODS ####################
    
    @classmethod
    def get_argument_parser(cls,install_or_manage) :
        parser = OpenMSIStreamArgumentParser()
        if install_or_manage=='install' :
            #subparsers from the classes that could be run
            subp_desc = 'The name of a runnable class to install as a service must be given as the first argument. '
            subp_desc = 'Adding one of these class names to the command line along with "-h" will show additional help.'
            parser.add_subparsers(description=subp_desc,required=True,dest='service_class_name')
            for service_dict in SERVICE_CONST.AVAILABLE_SERVICES :
                parser.add_subparser_arguments_from_class(service_dict['class'],addl_args=['optional_service_name'])
        elif install_or_manage=='manage' :
            parser.add_arguments('service_name','run_mode','remove_env_vars','remove_install_args','remove_nssm')
        else :
            errmsg =  'ERROR: must call get_argument_parser with either "install" or "manage", '
            errmsg+= f'not "{install_or_manage}"!'
            raise ValueError(errmsg)
        return parser

    #################### PROPERTIES ####################

    @property
    def env_var_names(self) :
        """
        Names of the environment variables used by the service
        """
        #get the names of environment variables from the command line and config file
        if self.argslist is not None and self.service_dict is not None :
            for arg in self.argslist :
                if arg.startswith('$') :
                    yield arg
            p = self.service_dict['class'].get_argument_parser()
            argsdests = [action.dest for action in p._actions]
            if 'config' in argsdests :
                pargs = p.parse_args(args=self.argslist)
                cfp = ConfigFileParser(pargs.config,logger=SERVICE_CONST.LOGGER)
                for evn in cfp.env_var_names :
                    yield evn

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,service_name,*args,service_class_name=None,argslist=None,interactive=None,**kwargs) :
        super().__init__(*args,**kwargs)
        self.service_name = service_name
        self.service_class_name = service_class_name
        self.__set_service_dict_from_class_name()
        self.argslist = argslist
        self.interactive = interactive
        self.env_var_filepath = SERVICE_CONST.WORKING_DIR/f'{self.service_name}_env_vars.txt'
        self.install_args_filepath = SERVICE_CONST.WORKING_DIR/f'{self.service_name}_install_args.txt'
        self.exec_fp = SERVICE_CONST.WORKING_DIR/f'{self.service_name}{SERVICE_CONST.SERVICE_EXECUTABLE_NAME_STEM}'
    
    def install_service(self) :
        """
        Install the Service
        child classes should call this method to get the setup done before they do anything specific
        """
        #make sure the necessary information was supplied
        if self.service_class_name is None or self.service_dict is None or self.argslist is None :
            errmsg = 'ERROR: newly installing a Service requires that the class name and argslist are supplied!'
            self.logger.error(errmsg,RuntimeError)
        #set the environment variables
        must_rerun = set_env_vars(self.env_var_names,interactive=self.interactive)
        if must_rerun :
            msg = 'New values for environment variables have been set. '
            msg+= 'Please close this window and rerun InstallService so that their values get picked up.'
            self.logger.info(msg)
            sys.exit(0)
        #write out the environment variable file
        self.env_vars_needed = self._write_env_var_file()
        #write out the install arguments file
        self._write_install_args_file()
        #write out the executable file
        self._write_executable_file()
        #install the service
        msg='Installing a'
        if self.service_class_name[0].lower() in ('a','e','i','o','u','y') :
            msg+='n'
        msg+=f' {self.service_class_name} program as "{self.service_name}" from executable at {self.exec_fp}...'
        self.logger.info(msg)

    def run_manage_command(self,run_mode,remove_env_vars=False,remove_install_args=False,remove_nssm=False) :
        """
        Run one of the functions below according to the action given
        """
        if run_mode in ['start'] :
            self.start_service()
        if run_mode in ['status'] :
            self.service_status()
        if run_mode in ['stop','stop_and_remove','stop_and_reinstall','stop_and_restart'] :
            self.stop_service()
        if run_mode in ['remove','stop_and_remove','stop_and_reinstall','stop_and_restart'] :
            self.remove_service(remove_env_vars=remove_env_vars,
                                remove_install_args=remove_install_args,
                                remove_nssm=remove_nssm)
        if run_mode in ['reinstall','stop_and_reinstall','stop_and_restart'] :
            self.reinstall_service()
        if run_mode in ['stop_and_restart'] :
            self.start_service()

    @abstractmethod
    def start_service(self) :
        """
        start the Service
        not implemented in the base class
        """
        pass

    @abstractmethod
    def service_status(self) :
        """
        get the status of the service
        not implemented in the base class
        """
        pass

    @abstractmethod
    def stop_service(self) :
        """
        stop the Service
        not implemented in the base class
        """
        pass

    def remove_service(self,remove_env_vars=False,remove_install_args=False) :
        """
        remove the Service
        child classes should call this after doing whatever they need to do to possibly remove environment variables
        """
        #remove the executable file
        if self.exec_fp.is_file() :
            self.exec_fp.unlink()
        #remove the environment variables that were set when the service was installed
        if remove_env_vars :
            try :
                for env_var_name in self.env_var_names :
                    remove_env_var(env_var_name)
                    self.logger.info(f'{env_var_name} environment variable successfully removed')
            except CalledProcessError :
                warnmsg = 'WARNING: failed to remove environment variables. You should remove any username/password '
                warnmsg+= 'environment variables manually even though the service is uninstalled!'
                self.logger.info(warnmsg)
        else :
            self.logger.info('Environment variables will be retained')
        #remove the actual environment variable file
        if self.env_var_filepath.is_file() :
            self.env_var_filepath.unlink()
        #remove the installation arguments file
        if remove_install_args :
            if self.install_args_filepath.is_file() :
                self.install_args_filepath.unlink()
                self.logger.info(f'Installation arguments file {self.install_args_filepath} deleted.')
        else :
            if self.install_args_filepath.is_file() :
                self.logger.info(f'Installation arguments file {self.install_args_filepath} will be retained')
        self.logger.info(f'Done removing {self.service_name}')
    
    def reinstall_service(self) :
        """
        Install the service using the arguments defined in the service arguments file
        """
        if not self.install_args_filepath.is_file() :
            errmsg =  'ERROR: cannot reinstall service without an installation arguments file '
            errmsg+= f'at {self.install_args_filepath}!'
            self.logger.error(errmsg,FileNotFoundError)
        with open(self.install_args_filepath,'r') as fp :
            lines = fp.readlines()
        if not len(lines)>0 :
            errmsg = f'ERROR: installation arguments file {self.install_args_filepath} does not contain enough entries!'
            self.logger.error(errmsg,RuntimeError)
        self.service_class_name = lines[0].strip()
        self.__set_service_dict_from_class_name()
        self.argslist = [line.strip() for line in lines[1:]] if len(lines)>1 else []
        if self.interactive :
            msg = 'Running this reinstall would be like running the following from the command line:\n'
            msg+= f'InstallService {self.service_class_name} '
            for arg in self.argslist :
                msg+=f'{arg} '
            msg+=f'--service_name {self.service_name}\n'
            msg+='Does everything above look right? [(y)/n]: '
            check = input(msg)
            if check.lower().startswith('n') :
                msg = f'{self.service_name} will NOT be reinstalled. You can install it again by rerunning '
                msg+= 'InstallService on its own.'
                self.logger.info(msg)
                sys.exit(0)
            else :
                self.install_service()

    #################### PRIVATE HELPER FUNCTIONS ####################
    
    @abstractmethod
    def _write_env_var_file(self) :
        """
        Write and set permissions for the file holding the values of the environment variables needed by the Service
        returns True if there are environment variables referenced, False if not
        not implemented in base class
        """
        pass

    def _write_install_args_file(self) :
        """
        Write out a file storing the arguments used to install the Service so it can be 
        reinstalled using the same configurations/setting if desired
        """
        if self.service_class_name is None or self.argslist is None :
            errmsg = "ERROR: can't write the installation arguments file without a class name and argslist!"
            self.logger.error(errmsg,RuntimeError)
        with open(self.install_args_filepath,'w') as fp :
            for arg in [self.service_class_name,*self.argslist] :
                fp.write(f'{arg}\n')
    
    def _write_executable_file(self,filepath=None) :
        """
        write out the executable python file that the service will actually be running
        """
        error_log_path = pathlib.Path().resolve()/f'{self.service_name}{SERVICE_CONST.ERROR_LOG_STEM}'
        code = f'''\
            if __name__=='__main__' :
                try :
                    from {self.service_dict['filepath']} import {self.service_dict['func_name']}
                    {self.service_dict['func_name']}({self.argslist})
                except Exception :
                    import pathlib, traceback, datetime
                    output_filepath = pathlib.Path(r"{error_log_path}")
                    with open(output_filepath,'a') as fp :'''
        code+=r'''
                        fp.write(f'{(datetime.datetime.now()).strftime("Error on %Y-%m-%d at %H:%M:%S")}. Exception:\n{traceback.format_exc()}')
                    import sys
                    sys.exit(1)
        '''
        if filepath is not None :
            warnmsg = f'WARNING: Services executable will be written to {filepath} instead of {self.exec_fp}'
            self.logger.warning(warnmsg)
            exec_fp = filepath
        else :
            exec_fp = self.exec_fp
        with open(exec_fp,'w') as fp :
            fp.write(textwrap.dedent(code))

    def __set_service_dict_from_class_name(self) :
        """
        If a service class name has been set, set the service dict with information about the code that should be run
        """
        if self.service_class_name is not None :
            #set the dictionary with details about the program that will run
            service_dict = [sd for sd in SERVICE_CONST.AVAILABLE_SERVICES if sd['script_name']==self.service_class_name]
            if len(service_dict)!=1 :
                errmsg = f'ERROR: could not find the Service dictionary for {self.service_name} '
                errmsg+= f'(a {self.service_class_name} program)! service_dict = {service_dict}'
                self.logger.error(errmsg,RuntimeError)
            self.service_dict = service_dict[0]
        else :
            self.service_dict = None



#imports
import pathlib, datetime
from typing import Set
from dataclasses import dataclass
from ..utilities.dataclass_table import DataclassTable
from ..utilities.logging import LogOwner

@dataclass
class RegistryLineInProgress :
    filename : str
    filepath : pathlib.Path
    n_chunks : int
    n_chunks_delivered : int
    n_chunks_to_send : int
    started : datetime.datetime
    chunks_delivered : Set[int]
    chunks_to_send : Set[int]

@dataclass
class RegistryLineCompleted :
    filename : str
    filepath : pathlib.Path
    n_chunks : int
    started : datetime.datetime
    completed : datetime.datetime

class ProducerFileRegistry(LogOwner) :
    """
    A class to manage two atomic csv files listing which portions 
    of which files have been uploaded to a particular topic
    """

    def __init__(self,dirpath,topic_name,*args,**kwargs) :
        """
        dirpath    = path to the directory that should contain the csv file
        topic_name = the name of the topic that will be produced to (used in the filename)
        """
        super().__init__(*args,**kwargs)
        #A table to list the files currently in progress
        in_prog_filepath = dirpath / f'files_to_upload_to_{topic_name}.csv'
        self.__in_prog = DataclassTable(dataclass_type=RegistryLineInProgress,filepath=in_prog_filepath,
                                        logger=self.logger)
        #A table to list the files that have been completely produced
        completed_filepath = dirpath / f'files_fully_uploaded_to_{topic_name}.csv'
        self.__completed = DataclassTable(dataclass_type=RegistryLineCompleted,filepath=completed_filepath,
                                          logger=self.logger)

    def get_incomplete_filepaths_and_chunks(self) :
        """
        Generate tuples of (filepath, chunks to upload) for each file that has not yet been completely uploaded
        """
        for obj_address in self.__in_prog.obj_addresses :
            attr_dict = self.__in_prog.get_entry_attrs(obj_address,'filepath','chunks_to_send')
            if len(attr_dict['chunks_to_send'])>0 :
                yield attr_dict['filepath'],attr_dict['chunks_to_send']

    def get_completed_filepaths(self) :
        """
        Generate filepaths for each file that has been completely uploaded
        """
        for obj_address in self.__completed.obj_addresses :
            yield self.__completed.get_entry_attrs(obj_address,'filepath')

    def register_chunk(self,filename,filepath,n_total_chunks,chunk_i) :
        """
        Register a chunk as having been successfully produced
        Returns True if all chunks for the file have been produced
        """
        #get a dictionary of the existing object addresses keyed by their filepaths
        existing_obj_addresses = self.__in_prog.obj_addresses_by_key_attr('filepath')
        #if the file is already recognized as in progress
        if filepath in existing_obj_addresses.keys() :
            #make sure there's only one object with this filepath
            if len(existing_obj_addresses[filepath])!=1 :
                errmsg = f'ERROR: found {len(existing_obj_addresses[filepath])} files in the producer registry '
                errmsg+= f'for filepath {filepath}'
                self.logger.error(errmsg,RuntimeError)
            existing_addr = existing_obj_addresses[filepath][0]
            #make sure the total numbers of chunks match
            existing_n_chunks = self.__in_prog.get_entry_attrs(existing_addr,'n_chunks')
            if existing_n_chunks!=n_total_chunks :
                errmsg = f'ERROR: {filepath} in {self.__class__.__name__} is listed as having '
                errmsg+= f'{existing_n_chunks} total chunks, but the producer callback for this file '
                errmsg+= f'lists {n_total_chunks} chunks! Did the chunk size change?'
                self.logger.error(errmsg,RuntimeError)
            #editing the object's attributes has to be done in one thread at a time
            with self.__in_prog.lock :
                #get its current state
                attrs = self.__in_prog.get_entry_attrs(existing_addr,'chunks_delivered','chunks_to_send')
                #if the chunk is already registered, just return
                if chunk_i in attrs['chunks_delivered'] :
                    return False
                #otherwise add/remove it from the sets
                attrs['chunks_delivered'].add(chunk_i)
                try :
                    attrs['chunks_to_send'].remove(chunk_i)
                except KeyError :
                    pass
                #reset the attributes for the object
                new_attrs = {
                    'n_chunks_delivered':len(attrs['chunks_delivered']),
                    'n_chunks_to_send':len(attrs['chunks_to_send']),
                    'chunks_delivered':attrs['chunks_delivered'],
                    'chunks_to_send':attrs['chunks_to_send'],
                    }
            self.__in_prog.set_entry_attrs(existing_addr,**new_attrs)
            # if there are no more chunks to send and all chunks have been delivered, 
            # move this file from "in progress" to "completed" and force-dump the files
            if ( self.__in_prog.get_entry_attrs(existing_addr,'n_chunks_delivered')==n_total_chunks and 
                 self.__in_prog.get_entry_attrs(existing_addr,'n_chunks_to_send')==0 ) :
                started = self.__in_prog.get_entry_attrs(existing_addr,'started')
                completed_entry = RegistryLineCompleted(filename,filepath,n_total_chunks,
                                                        started,datetime.datetime.now())
                self.__completed.add_entries(completed_entry)
                self.__completed.dump_to_file()
                self.__in_prog.remove_entries(existing_addr)
                self.__in_prog.dump_to_file()
                return True
            else :
                return False
        #otherwise it's a new file to list somewhere
        else :
            to_deliver = set([])
            for ic in range(1,n_total_chunks+1) :
                if ic!=chunk_i :
                    to_deliver.add(ic)
            #if there are other chunks to deliver, register it to "in_progress"
            if len(to_deliver)>0 :
                in_prog_entry = RegistryLineInProgress(filename,filepath,n_total_chunks,
                                                       1,n_total_chunks-1,datetime.datetime.now(),
                                                       set([chunk_i]),to_deliver)
                self.__in_prog.add_entries(in_prog_entry)
                self.__in_prog.dump_to_file()
                return False
            #otherwise register it as a completed file
            else :
                completed_entry = RegistryLineCompleted(filename,filepath,n_total_chunks,
                                                        datetime.datetime.now(),datetime.datetime.now())
                self.__completed.add_entries(completed_entry)
                self.__completed.dump_to_file()
                return True



#imports
import pathlib, math, re
from argparse import ArgumentParser
from ..data_file_io.config import RUN_OPT_CONST
from .config import RUN_CONST

#################### MISC. FUNCTIONS ####################

#convert a string or path argument into a path to a file, checking if it exists
def existing_file(argstring) :
    filepath = pathlib.Path(argstring)
    if filepath.is_file() :
        return filepath.resolve()
    raise FileNotFoundError(f'ERROR: file {argstring} does not exist!')

#convert a string or path argument into a directory path, checking if it exists
def existing_dir(argstring) :
    dirpath = pathlib.Path(argstring)
    if dirpath.is_dir() :
        return dirpath.resolve()
    raise FileNotFoundError(f'ERROR: directory {argstring} does not exist!')

#convert a string or path argument into a directory path, creating it if necessary
def create_dir(argstring) :
    if argstring is None : #Then the argument wasn't given and nothing should be done
        return None
    dirpath = pathlib.Path(argstring)
    if dirpath.is_dir() :
        return dirpath.resolve()
    try :
        dirpath.mkdir(parents=True)
        return dirpath.resolve()
    except Exception as e :
        raise RuntimeError(f'ERROR: failed to create directory with name {argstring}! error: {e}')

#convert a string or path argument into a config file path (raise an exception if the file can't be found)
def config_path(configarg) :
    if isinstance(configarg,str) and '.' not in configarg :
        configarg+=RUN_CONST.CONFIG_FILE_EXT
    configpath = pathlib.Path(configarg)
    if configpath.is_file() :
        return configpath.resolve()
    if (RUN_CONST.CONFIG_FILE_DIR/configpath).is_file() :
        return (RUN_CONST.CONFIG_FILE_DIR/configpath).resolve()
    raise ValueError(f'ERROR: config argument {configarg} is not a recognized config file!')

#detects if the bucket name contains invalid characters
def detect_bucket_name(argstring) :
    if argstring is None : #Then the argument wasn't given and nothing should be done
        return None
    illegal_charcters = ['#', '%', '&' ,'{', '}', '\\', '/', '<', '>', '*', '?', ' ', 
                         '$', '!', '\'', '\"', ':', '@', '+', '`', '|', '=']
    if argstring in illegal_charcters:
        raise RuntimeError(f'ERROR: Illegal characters in bucket_name {argstring}')
    return argstring

#make sure a given value is a nonzero integer power of two (or can be converted to one)
def int_power_of_two(argval) :
    if not isinstance(argval,int) :
        try :
            argval=int(argval)
        except Exception as e :
            raise ValueError(f'ERROR: could not convert {argval} to an integer in int_power_of_two! Exception: {e}')
    if argval<=0 or math.ceil(math.log2(argval))!=math.floor(math.log2(argval)) :
        raise ValueError(f'ERROR: invalid argument: {argval} must be a (nonzero) power of two!')
    return argval

#make sure a given value is a positive integer
def positive_int(argval) :
    try :
        argval = int(argval)
    except Exception as e :
        raise ValueError(f'ERROR: could not convert {argval} to an integer in positive_int! Exception: {e}')
    if (not isinstance(argval,int)) or (argval<1) :
        raise ValueError(f'ERROR: invalid argument: {argval} must be a positive integer!')
    return argval

#################### MYARGUMENTPARSER CLASS ####################

class OpenMSIStreamArgumentParser(ArgumentParser) :
    """
    Class to make it easier to get an ArgumentParser with some commonly-used arguments in it
    """

    ARGUMENTS = {
        'filepath':
            ['positional',{'type':existing_file,'help':'Path to the data file to use'}],
        'output_dir':
            ['positional',{'type':create_dir,'help':'Path to the directory to put output in'}],
        'upload_dir':
            ['positional',{'type':existing_dir,'help':'Path to the directory to watch for files to upload'}],
        'bucket_name':
            ['positional', {'type': detect_bucket_name, 
                            'help': '''Name of the object store bucket to which consumed files should be transferred.
                                       Name should only contain valid characters'''}],
        'config':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_CONFIG_FILE,'type':config_path,
                         'help':f'''Name of config file to use in {RUN_CONST.CONFIG_FILE_DIR.resolve()}, 
                                    or path to a file in a different location'''}],
        'topic_name':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_TOPIC_NAME,
                         'help':'Name of the topic to produce to or consume from'}],
        'logger_file':
            ['optional', {'default': pathlib.Path(), 'type': pathlib.Path,
                          'help': '''Path to the log file (or directory to hold the auto-named logfile)'''}],        
        'n_threads':
            ['optional',{'default':RUN_CONST.DEFAULT_N_THREADS,'type':positive_int,
                         'help':'Maximum number of threads to use'}],
        'upload_regex':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_UPLOAD_REGEX,'type':re.compile,
                         'help':'Only files matching this regular expression will be uploaded'}],
        'chunk_size':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_CHUNK_SIZE,'type':int_power_of_two,
                         'help':'''Max size (in bytes) of chunks into which files should be broken 
                                   as they are uploaded'''}],
        'queue_max_size':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_MAX_UPLOAD_QUEUE_SIZE,'type':int,
                         'help':'''Maximum number of items to allow in the upload queue at a time. 
                                 Use to limit RAM usage or throttle production rate if necessary.'''}],
        'update_seconds':
            ['optional',{'default':RUN_CONST.DEFAULT_UPDATE_SECONDS,'type':int,
                         'help':'''Number of seconds between printing a "." to the console 
                                   to indicate the program is alive'''}],
        'upload_existing':
            ['optional',{'action':'store_true',
                         'help':'''Add this flag to upload files already existing in addition to those 
                                   added to the directory after this code starts running (by default 
                                   only files added after startup will be uploaded)'''}],
        'consumer_group_ID':
            ['optional',{'default':'create_new',
                         'help':'ID to use for all consumers in the group'}],
        'optional_output_dir':
            ['optional',{'type':create_dir,
                         'help':'Optional path to directory to put output in'}],
        'service_name':
            ['positional',{'help':'The name of the service to work with'}],
        'optional_service_name':
            ['optional',{'help':'The customized name of the Service that will be installed to run the chosen class'}],
        'run_mode':
            ['positional',{'choices':['start','status','stop','remove','stop_and_remove',
                                      'reinstall','stop_and_reinstall','stop_and_restart'],
                           'help':'What to do with the service'}],
        'remove_env_vars':
            ['optional',{'action':'store_true',
                         'help':'''Add this flag to also remove username/password environment variables 
                                   when removing a Service'''}],
        'remove_install_args':
            ['optional',{'action':'store_true',
                         'help':'Add this flag to also remove the install arguments file when removing a Service'}],
        'remove_nssm':
            ['optional',{'action':'store_true',
                         'help':'Add this flag to also remove the NSSM executable when removing a Service'}],
    }

    #################### OVERLOADED FUNCTIONS ####################

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.__argnames_added = []
        self.__subparsers_action_obj = None
        self.__subparsers = {}
        self.__subparser_argnames_added = {}

    def add_subparsers(self,*args,**kwargs) :
        """
        Overloaded from base class; OpenMSIStreamArgumentParser actually owns its subparsers
        """
        if self.__subparsers_action_obj is not None or self.__subparsers!={} :
            raise RuntimeError('ERROR: add_subparsers called for an argument parser that already has subparsers!')
        self.__subparsers_action_obj = super().add_subparsers(*args,**kwargs)

    def parse_args(self,**kwargs) :
        """
        Overloaded from base class to print usage when catching errors in parsing arguments
        """
        try :
            return super().parse_args(**kwargs)
        except Exception :
            self.print_usage()
            raise

    #################### UNIQUE FUNCTIONS ####################

    def add_arguments(self,*args,**kwargs) :
        """
        Add a group of common arguments to the parser
        args = names of arguments that should be added just as listed in ARGUMENTS above
        kwargs = dictionary whose keys are argument names like in args and whose values are default argument values
        """
        if len(args)<1 and len(kwargs)<1 :
            raise ValueError('ERROR: must specify at least one desired argument to create an argument parser!')
        args_to_use = []
        for argname in args :
            if argname not in args_to_use :
                args_to_use.append(argname)
        for argname in kwargs.keys() :
            if argname in args_to_use :
                args_to_use.remove(argname)
        for argname in args_to_use :
            argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname)
            if argname_to_add not in self.__argnames_added :
                self.add_argument(argname_to_add,**kwargs_for_arg)
                self.__argnames_added.append(argname_to_add)
        for argname,argdefault in kwargs.items() :
            argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname,argdefault)
            if argname_to_add not in self.__argnames_added :
                self.add_argument(argname_to_add,**kwargs_for_arg)
                self.__argnames_added.append(argname_to_add)

    def add_subparser(self,cmd,**kwargs) :
        """
        Add a subparser by calling add_parser using the subparser action object
        Returns the subparser
        Just a wrapper around the usual UI
        """
        if self.__subparsers_action_obj is None :
            errmsg = 'ERROR: add_subparser_arguments_for_class called for an argument parser that '
            errmsg+= 'has not added subparsers!'
            raise RuntimeError(errmsg)
        self.__subparsers[cmd] = self.__subparsers_action_obj.add_parser(cmd,**kwargs)
        return self.__subparsers[cmd]

    def add_subparser_arguments_from_class(self,class_to_add,*,addl_args=None,addl_kwargs=None,**other_kwargs) :
        """
        Create a new subparser and add arguments from the given class to it
        class_to_add must inherit from Runnable to be able to get its arguments
        addl_args/kwargs = a list and/or dictionary of additional arguments that should be added to every subparser 
                                 (same format as args/kwargs for add_arguments above)
        other_kwargs are passed to subparsers.add_parser
        """
        if self.__subparsers_action_obj is None :
            errmsg = 'ERROR: add_subparser_arguments_for_class called for an argument parser that '
            errmsg+= 'has not added subparsers!'
            raise RuntimeError(errmsg)
        class_name = class_to_add.__name__
        if class_name in self.__subparsers.keys() :
            errmsg = f'ERROR: subparser arguments for {class_name} have already been added to this argument parser!'
            raise RuntimeError(errmsg)
        self.__subparsers[class_name] = self.__subparsers_action_obj.add_parser(class_name,**other_kwargs)
        self.__subparser_argnames_added[class_name] = []
        argnames, argnames_with_defaults = class_to_add.get_command_line_arguments()
        if addl_args is not None :
            argnames = [*argnames,*addl_args]
        if addl_kwargs is not None :
            argnames_with_defaults = {**argnames_with_defaults,**addl_kwargs}
        for argname in argnames :
            argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname)
            if argname_to_add not in self.__subparser_argnames_added[class_name] :
                self.__subparsers[class_name].add_argument(argname_to_add,**kwargs_for_arg)
                self.__subparser_argnames_added[class_name].append(argname_to_add)
        for argname,argdefault in argnames_with_defaults.items() :
            argname_to_add,kwargs_for_arg = self.__get_argname_and_kwargs(argname,argdefault)
            if argname_to_add not in self.__subparser_argnames_added[class_name] :
                self.__subparsers[class_name].add_argument(argname_to_add,**kwargs_for_arg)
                self.__subparser_argnames_added[class_name].append(argname_to_add)

    def __get_argname_and_kwargs(self,argname,new_default=None) :
        """
        Return the name and kwargs dict for a particular argument

        argname = the given name of the argument, will be matched to a name in ARGUMENTS
        new_default = an optional parameter given to override the default already specified in ARGUMENTS
        """
        if argname in self.ARGUMENTS.keys() :
            if self.ARGUMENTS[argname][0]=='positional' :
                argname_to_add = argname 
            else :
                if argname.startswith('optional_') :
                    argname_to_add = f'--{argname[len("optional_"):]}'
                else :
                    argname_to_add = f'--{argname}'
            kwargs = self.ARGUMENTS[argname][1].copy()
            if new_default is not None :
                if ( 'default' in kwargs.keys() and 
                    type(new_default)!=type(kwargs['default']) ) :
                    errmsg = f'ERROR: new default value {new_default} for argument {argname} is of a different type '
                    errmsg+= f'than expected based on the old default ({self.ARGUMENTS[argname]["kwargs"]["default"]})!'
                    raise ValueError(errmsg)
                kwargs['default'] = new_default
            if 'default' in kwargs.keys() :
                if 'help' in kwargs.keys() :
                    kwargs['help']+=f" (default = {kwargs['default']})"
                else :
                    kwargs['help']=f"default = {kwargs['default']}"
            return argname_to_add, kwargs
        else :
            raise ValueError(f'ERROR: argument {argname} is not recognized as an option!')



#imports
import pathlib, logging, traceback

class OpenMSIStreamFormatter(logging.Formatter) :
    """
    Very small extension of the usual logging.Formatter to allow modification of format based on message content
    """

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)

    def format(self, record):
        """
        If a message starts with a newline, start the actual logging line with the newline before any of the rest
        """
        formatted = ''
        if record.msg.startswith('\n') :
            record.msg = record.msg.lstrip('\n')
            formatted+='\n'
        formatted+=super().format(record)
        return formatted

class Logger :
    """
    Class for a general logger. Logs messages and raises exceptions
    """

    @property
    def formatter(self):
        return OpenMSIStreamFormatter('[%(name)s at %(asctime)s] %(message)s','%Y-%m-%d %H:%M:%S')

    def __init__(self,logger_name=None,streamlevel=logging.DEBUG,logger_filepath=None,filelevel=logging.INFO) :
        """
        name = the name for this logger to use (probably something like the top module that owns it)
        """
        self._name = logger_name
        if self._name is None :
            self._name = self.__name__
        self._logger_obj = logging.getLogger(self._name)
        self._logger_obj.setLevel(logging.DEBUG)
        self._streamhandler = logging.StreamHandler()
        self._streamhandler.setLevel(streamlevel)
        self._streamhandler.setFormatter(self.formatter)
        self._logger_obj.addHandler(self._streamhandler)
        self._filehandler = None
        if logger_filepath is not None :
            self.add_file_handler(logger_filepath)

    #set the level of the underlying logger
    def set_level(self,level) :
        self._logger_obj.setLevel(level)
    #set the level of the streamhandler
    def set_stream_level(self,level) :
        self._streamhandler.setLevel(level)
    #set the level of the filehandler
    def set_file_level(self,level) :
        if self._filehandler is None :
            errmsg = f'ERROR: Logger {self._name} does not have a filehandler set but set_file_level was called!'
            raise RuntimeError(errmsg)
        self._filehandler.setLevel(level)

    #add a filehandler to the logger
    def add_file_handler(self,filepath,level=logging.INFO) :
        if not isinstance(filepath,pathlib.PurePath) :
            self.error(f'ERROR: {filepath} is a {type(filepath)} object, not a Path object!',TypeError)
        if not filepath.is_file() :
            if not filepath.parent.is_dir() :
                filepath.parent.mkdir(parents=True)
            filepath.touch()
        self._filehandler = logging.FileHandler(filepath)
        self._filehandler.setLevel(level)
        self._filehandler.setFormatter(self.formatter)
        self._logger_obj.addHandler(self._filehandler)

    #methods for logging different levels of messages

    def debug(self,msg,*args,**kwargs) :
        self._logger_obj.debug(msg,*args,**kwargs)
    
    def info(self,msg,*args,**kwargs) :
        self._logger_obj.info(msg,*args,**kwargs)
    
    def warning(self,msg,*args,**kwargs) :
        if not msg.startswith('WARNING:') :
            msg = f'WARNING: {msg}'
        self._logger_obj.warning(msg,*args,**kwargs)

    #log an error message and optionally raise an exception with the same message, or raise a different exception
    def error(self,msg,exception_type=None,exc_obj=None,reraise=True,*args,**kwargs) :
        if not msg.startswith('ERROR:') :
            msg = f'ERROR: {msg}'
        self._logger_obj.error(msg,*args,**kwargs)
        if exc_obj is not None :
            self.log_exception_as_error(exc_obj,*args,reraise=reraise,**kwargs)
        if exception_type is not None : 
            raise exception_type(msg)

    #log the traceback of a given Exception as an error and optionally reraise it
    def log_exception_as_error(self,exc,*args,reraise=True,**kwargs) :
        try :
            raise exc
        except Exception :
            exc_to_log = 'ERROR: Exception: ['
            for line in (traceback.format_exc()).split('\n') :
                exc_to_log+=f'{line},'
            exc_to_log+=']'
            self._logger_obj.error(exc_to_log,*args,**kwargs)
            if reraise :
                raise exc

class LogOwner :
    """
    Any subclasses extending this one will have access to a Logger defined by the first class in the MRO to extend it
    """

    @property
    def logger(self) :
        return self.__logger

    @logger.setter
    def logger(self,logger) :
        if hasattr(self,'_LogOwner__logger') and self.__logger is not None and type(self.__logger)!=type(logger) :
            errmsg = f'ERROR: tried to reset a logger of type {type(self.__logger)} to a new logger of type {logger}!'
            self.__logger.error(errmsg,ValueError)
        else :
            self.__logger = logger

    def __init__(self,*args,
                 logger=None,logger_name=None,streamlevel=logging.DEBUG,logger_file=None,filelevel=logging.INFO,
                 **other_kwargs) :
        if logger is not None :
            self.__logger = logger
        else :
            if logger_name is None :
                logger_name = self.__class__.__name__
            logger_filepath = logger_file
            if logger_file is not None and logger_file.is_dir() :
                logger_filepath = logger_file / f'{self.__class__.__name__}.log'
            self.__logger = Logger(logger_name,streamlevel,logger_filepath,filelevel)
        super().__init__(*args,**other_kwargs)



#imports
import traceback
from abc import ABC, abstractmethod
from kafkacrypto.message import KafkaCryptoMessage
from ..utilities.logging import LogOwner
from .config import DATA_FILE_HANDLING_CONST
from .utilities import get_encrypted_message_timestamp_string
from .download_data_file import DownloadDataFileToMemory
from .data_file_chunk_processor import DataFileChunkProcessor

class DataFileStreamProcessor(DataFileChunkProcessor,LogOwner,ABC) :
    """
    A class to consume DataFileChunk messages into memory and perform some operation(s) when entire files are available
    """

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,datafile_type=DownloadDataFileToMemory,**kwargs) :
        """
        datafile_type = the type of datafile that the consumed messages should be assumed to represent
        In this class datafile_type should be something that extends DownloadDataFileToMemory
        """   
        super().__init__(*args,datafile_type=datafile_type,**kwargs)
        if not issubclass(datafile_type,DownloadDataFileToMemory) :
            errmsg = 'ERROR: DataFileStreamProcessor requires a datafile_type that is a subclass of '
            errmsg+= f'DownloadDataFileToMemory but {datafile_type} was given!'
            self.logger.error(errmsg,ValueError)

    def process_files_as_read(self) :
        """
        Consumes messages and stores their data together separated by their original files.
        Uses several parallel threads to consume message and process fully read files. 
        Returns the total number of messages read and a list of the fully processed filenames.
        """
        msg = f'Will process files from messages in the {self.topic_name} topic using {self.n_threads} '
        msg+= f'thread{"s" if self.n_threads>1 else ""}'
        self.logger.info(msg)
        self.run()
        return self.n_msgs_read, self.n_msgs_processed, self.completely_processed_filepaths

    #################### PRIVATE HELPER FUNCTIONS ####################

    @abstractmethod
    def _process_downloaded_data_file(self,datafile,lock) :
        """
        Perform some operations on a given data file that has been fully read from the stream
        Can optionally lock other threads using the given lock
        Returns None if processing was successful and an Exception otherwise
        Not implemented in the base class
        """
        pass

    def _undecryptable_message_callback(self,msg) :
        """
        This function is called when a message that could not be decrypted in found.
        It should return True if an undecrypted message is considered "successfully processed" and False otherwise
        If this function is called it is likely that the file the chunk is coming from won't be able to be processed.

        In the base class, this logs a warning and returns False. 
        Overload this in child classes to do something more sensible.

        msg = the Message object from KafkaCrypto with KafkaCryptoMessages for its key and/or value
        """
        timestamp_string = get_encrypted_message_timestamp_string(msg)
        warnmsg = f'WARNING: encountered a message that failed to be decrypted (timestamp = {timestamp_string}). '
        warnmsg+= 'This message will be skipped, and the file it came from likely cannot be processed from the stream.'
        self.logger.warning(warnmsg)
        return False

    def _process_message(self,lock,msg):
        retval = super()._process_message(lock, msg, None, self.logger)
        #if the message was returned because it couldn't be decrypted, write it to the encrypted messages directory
        if ( hasattr(retval,'key') and hasattr(retval,'value') and 
             (isinstance(retval.key,KafkaCryptoMessage) or isinstance(retval.value,KafkaCryptoMessage)) ) :
            return self._undecryptable_message_callback(retval)
        if retval==True :
            return retval
        #get the DataFileChunk from the message value
        try :
            dfc = msg.value() #from a regular Kafka Consumer
        except :
            dfc = msg.value #from KafkaCrypto
        #if the file has had all of its messages read successfully, send it to the processing function
        if retval==DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE :
            if dfc.rootdir is not None :
                short_filepath = self.files_in_progress_by_path[dfc.filepath].full_filepath.relative_to(dfc.rootdir)
            else :
                short_filepath = self.files_in_progress_by_path[dfc.filepath].filepath
            msg = f'Processing {short_filepath}...'
            self.logger.info(msg)
            processing_retval = self._process_downloaded_data_file(self.files_in_progress_by_path[dfc.filepath],lock)
            with lock :
                del self.files_in_progress_by_path[dfc.filepath]
                del self.locks_by_fp[dfc.filepath]
            #if it was able to be processed
            if processing_retval is None :
                self.logger.info(f'Fully-read file {short_filepath} successfully processed')
                self.completely_processed_filepaths.append(dfc.filepath)
                return True
            #warn if it wasn't processed correctly
            else :
                if isinstance(processing_retval,Exception) :
                    try :
                        raise processing_retval
                    except Exception :
                        self.logger.info(traceback.format_exc())
                else :
                    self.logger.error(f'Return value from _process_downloaded_data_file = {processing_retval}')
                warnmsg = f'WARNING: Fully-read file {short_filepath} was not able to be processed. '
                warnmsg+= 'Check log lines above for more details on the specific error. '
                warnmsg+= 'The messages for this file will need to be consumed again if the file is to be processed!'
                self.logger.warning(warnmsg)
                return False
        #if the file hashes didn't match, return False
        elif retval==DATA_FILE_HANDLING_CONST.FILE_HASH_MISMATCH_CODE :
            warnmsg = f'WARNING: file hashes for file {self.files_in_progress_by_path[dfc.filepath].filename} '
            warnmsg+= 'not matched after being fully read! This file will not be processed.'
            self.logger.warning(warnmsg)
            with lock :
                del self.files_in_progress_by_path[dfc.filepath]
                del self.locks_by_fp[dfc.filepath]
            return False
        else :
            self.logger.error(f'ERROR: unrecognized add_chunk return value ({retval})!',NotImplementedError)
            return False

    def _on_check(self) :
        msg = f'{self.n_msgs_read} messages read, {self.n_msgs_processed} messages processed, '
        msg+= f'{len(self.completely_processed_filepaths)} files completely processed so far'
        self.logger.debug(msg)
        if len(self.files_in_progress_by_path)>0 or len(self.completely_processed_filepaths)>0 :
            self.logger.debug(self.progress_msg)



#imports
from abc import ABC, abstractmethod
from .argument_parsing import OpenMSIStreamArgumentParser
from .has_argument_parser import HasArgumentParser

class Runnable(HasArgumentParser,ABC) :
    """
    Class for any child classes that can be run on their own from the command line
    """

    ARGUMENT_PARSER_TYPE = OpenMSIStreamArgumentParser

    @classmethod
    @abstractmethod
    def get_command_line_arguments(cls) :
        """
        child classes should implement this function to return a list of argument names 
        and a dictionary of argument names/default values to add to the argument parser
        """
        return [],{}

    @classmethod
    def get_argument_parser(cls,*args,**kwargs) :
        """
        Return the argument parser used to run the code
        """
        parser = cls.ARGUMENT_PARSER_TYPE(*args,**kwargs)
        cl_args, cl_kwargs = cls.get_command_line_arguments()
        parser.add_arguments(*cl_args,**cl_kwargs)
        return parser

    @classmethod
    @abstractmethod
    def run_from_command_line(cls,args=None) :
        """
        child classes should implement this function to do whatever it is they do when they run from the command line
        """
        pass 



#imports
from ..utilities.misc import populated_kwargs
from ..utilities.logging import LogOwner

class DataFile(LogOwner) :
    """
    Base class for representing a single data file
    """

    #################### PROPERTIES ####################

    @property
    def filepath(self):
        return self.__filepath
    @property
    def filename(self):
        return self.__filename
    @filename.setter
    def filename(self,fn) :
        self.__filename = fn

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,filepath,*args,**kwargs) :
        """
        filepath = path to the file
        """
        self.__filepath = filepath
        self.__filename = self.__filepath.name
        kwargs = populated_kwargs(kwargs,{'logger_file':self.__filepath.parent})
        super().__init__(*args,**kwargs)




#imports
from threading import Lock
from abc import ABC, abstractmethod
from ..utilities.exception_tracking_thread import ExceptionTrackingThread
from .config import RUN_CONST
from .controlled_process import ControlledProcess

class ControlledProcessMultiThreaded(ControlledProcess,ABC) :
    """
    A class for running a group of processes in multiple threads until they're explicitly shut down
    """

    @property
    def n_threads(self):
        return self.__n_threads

    def __init__(self,*args,n_threads=RUN_CONST.DEFAULT_N_THREADS,**kwargs) :
        """
        n_threads = number of threads to use
        """
        self.__n_threads = n_threads
        super().__init__(*args,**kwargs)
        self.__lock = Lock()

    def run(self,args_per_thread=[],kwargs_per_thread={}) :
        """
        args_per_thread = a list of lists of arguments that should be given to the independent threads
                          one list of arguments per thread
        kwargs_per_thread = a list of dicts of keyword arguments that should be given to the independent threads
                          one dict of keyword arguments per thread
        """
        super().run()
        #correct the arguments for each thread
        self.__args_per_thread = args_per_thread
        if self.__args_per_thread==[] or (not type(self.__args_per_thread)==list) :
            self.__args_per_thread = [self.__args_per_thread]
        if not len(self.__args_per_thread)==self.__n_threads :
            if not len(self.__args_per_thread)==1 :
                errmsg = 'ERROR: ControlledProcessMultiThreaded.run was given a list of arguments with '
                errmsg+= f'{len(self.__args_per_thread)} entries, but was set up to use {self.__n_threads} threads!'
                self.logger.error(errmsg,ValueError)
            else :
                self.__args_per_thread = self.__n_threads*self.__args_per_thread
        #correct the keyword arguments for each thread
        self.__kwargs_per_thread = kwargs_per_thread
        if not type(self.__kwargs_per_thread)==list :
            self.__kwargs_per_thread = [self.__kwargs_per_thread]
        if not len(self.__kwargs_per_thread)==self.__n_threads :
            if not len(self.__kwargs_per_thread)==1 :
                errmsg = 'ERROR: ControlledProcessMultiThreaded.run was given a list of arguments with '
                errmsg+= f'{len(self.__kwargs_per_thread)} entries, but was set up to use {self.__n_threads} threads!'
                self.logger.error(errmsg,ValueError)
            else :
                self.__kwargs_per_thread = self.__n_threads*self.__kwargs_per_thread
        #create and start the independent threads
        self.__threads = []
        for i in range(self.__n_threads) :
            self.__threads.append(ExceptionTrackingThread(target=self._run_worker,
                                           args=(self.__lock,*self.__args_per_thread[i]),
                                           kwargs=self.__kwargs_per_thread[i]))
            self.__threads[-1].start()
        #loop while the process is alive, checking the control command queue and printing the "still alive" character
        while self.alive :
            self._print_still_alive()
            self._check_control_command_queue()
            self.__restart_crashed_threads()

    def _on_shutdown(self) :
        """
        Can override this method further in subclasses, just be sure to also call super()._on_shutdown() to join threads
        """
        for t in self.__threads :
            t.join()

    @abstractmethod
    def _run_worker(self,lock,*args,**kwargs) :
        """
        A function that should include a while self.alive loop for each independent thread to run 
        until the process is shut down

        lock = a Lock that can be used across all shared threads to guaranteee exactly one process is running

        Not implemented in the base class
        """
        pass

    def __restart_crashed_threads(self) :
        """
        Log Exceptions thrown by any of the threads and restart them
        """
        for ti,thread in enumerate(self.__threads) :
            if thread.caught_exception is not None :
                #log the error
                warnmsg = 'WARNING: a thread raised an Exception, which will be logged as an error below but not '
                warnmsg+= 'reraised. The thread that raised the error will be restarted.'
                self.logger.warning(warnmsg)
                self.logger.log_exception_as_error(thread.caught_exception,reraise=False)
                #try to join the thread 
                try :
                    thread.join()
                except :
                    pass
                finally :
                    self.__threads[ti] = None
                #restart the thread
                self.__threads[ti] = ExceptionTrackingThread(target=self._run_worker,
                                              args=(self.__lock,*self.__args_per_thread[ti]),
                                              kwargs=self.__kwargs_per_thread[ti])
                self.__threads[ti].start()



#imports
import pathlib, uuid
from kafkacrypto import KafkaProducer, KafkaConsumer, KafkaCrypto
from ..utilities.misc import cd

class OpenMSIStreamKafkaCrypto :
    """
    A class to own and work with the Producers, Consumers, and other objects needed
    by KafkaCrypto to handle encrypting/decrypting messages and exchanging keys
    """

    @property
    def key_serializer(self) :
        return self._kc.getKeySerializer()
    @property
    def key_deserializer(self) :
        return self._kc.getKeyDeserializer()
    @property
    def value_serializer(self) :
        return self._kc.getValueSerializer()
    @property
    def value_deserializer(self) :
        return self._kc.getValueDeserializer()

    def __init__(self,server_configs,config_file) :
        """
        server_configs = the producer/consumer-agnostic configurations for connecting to the server in question
        config_file    = the path to the config file that should be used to instantiate KafkaCrypto 
        """
        producer_configs = server_configs.copy()
        consumer_configs = server_configs.copy()
        #figure out a consumer group ID to use (KafkaCrypto Consumers need one)
        if 'group.id' not in consumer_configs.keys() :
            consumer_configs['group.id'] = str(uuid.uuid1())
        with cd(pathlib.Path(config_file).parent) :
            #start up the producer and consumer
            self._kcp = KafkaProducer(**producer_configs)
            self._kcc = KafkaConsumer(**consumer_configs)
            #initialize the KafkaCrypto object 
            self._kc = KafkaCrypto(None,self._kcp,self._kcc,config_file)

    def close(self) :
        try :
            self._kc.close()
        except :
            pass
        finally :
            self._kc = None
            self._kcp = None
            self._kcc = None



#imports
import traceback, time
from threading import Thread
from queue import Queue
from hashlib import sha512
from ..running.runnable import Runnable
from ..kafka_wrapper.producer_group import ProducerGroup
from .config import RUN_OPT_CONST
from .data_file_chunk import DataFileChunk
from .data_file import DataFile

class UploadDataFile(DataFile,Runnable) :
    """
    Class to represent a data file whose messages will be uploaded to a topic
    """

    #################### PROPERTIES ####################

    @property
    def select_bytes(self) :
        return []   # in child classes this can be a list of tuples of (start_byte,stop_byte) 
                    # in the file that will be the only ranges of bytes added when creating the list of chunks
    @property
    def rootdir(self) :
        return self.__rootdir
    @property
    def chunks_to_upload(self) :
        return self.__chunks_to_upload
    @property
    def to_upload(self):
        return self.__to_upload #whether or not this file will be considered when uploading some group of data files
    @to_upload.setter
    def to_upload(self,tu) :
        if tu and (not self.__to_upload) :
            self.logger.debug(f'Setting {self.filepath} to be uploaded')
        elif (not tu) and self.__to_upload :
            self.logger.debug(f'Setting {self.filepath} to NOT be uploaded')
        self.__to_upload = tu
    @property
    def fully_produced(self) : #whether or not this file has had all of its chunks successfully sent to the broker
        return self.__fully_produced
    @fully_produced.setter
    def fully_produced(self,fp) :
        self.__fully_produced = fp
    @property
    def fully_enqueued(self) : #whether or not this file has had all of its chunks added to an upload queue somewhere
        return self.__fully_enqueued
    @property
    def waiting_to_upload(self) : #whether or not this file is waiting for its upload to begin
        if (not self.__to_upload) or self.__fully_enqueued :
            return False
        if len(self.__chunks_to_upload)>0 :
            return False
        return True
    @property
    def upload_in_progress(self) : #whether this file is in the process of being enqueued to be uploaded
        if (not self.__to_upload) or self.__fully_enqueued :
            return False
        if len(self.__chunks_to_upload)==0 :
            return False
        return True
    @property
    def upload_status_msg(self) : #a message stating the file's name and status w.r.t. being enqueued to be uploaded 
        if self.__rootdir is None :
            msg = f'{self.filepath} '
        else :
            msg = f'{self.filepath.relative_to(self.__rootdir)} '
        if not self.__to_upload :
            msg+='(will not be uploaded)'
        elif self.__fully_produced :
            msg+=f'({self.__n_total_chunks} message'
            if self.__n_total_chunks!=1 :
                msg+='s'
            msg+=' fully produced to broker)'
        elif self.__fully_enqueued :
            msg+=f'({self.__n_total_chunks} message'
            if self.__n_total_chunks!=1 :
                msg+='s'
            msg+=' fully enqueued)'
        elif self.upload_in_progress :
            msg+=f'(in progress with {self.__n_total_chunks-len(self.__chunks_to_upload)}'
            msg+=f'/{self.__n_total_chunks} messages enqueued)'
        elif self.waiting_to_upload :
            msg+=f'({self.__n_total_chunks} messages waiting to be enqueued)'
        else :
            msg+='(status unknown)'
        return msg

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,to_upload=True,rootdir=None,filename_append='',**kwargs) :
        """
        to_upload       = if False, the file will be ignored for purposes of uploading to a topic (default is True)
        rootdir         = path to the "root" directory that this file is in; anything in the path beyond it 
                          will be added to the DataFileChunk so that it will be reconstructed inside a subdirectory
        filename_append = a string that should be appended to the end of the filename stem to distinguish the file 
                          that's produced from its original file on disk
        """
        super().__init__(*args,**kwargs)
        self.__to_upload = to_upload
        if rootdir is None :
            self.__rootdir = self.filepath.parent
        else :
            self.__rootdir = rootdir
        self.__filename_append = filename_append
        self.__fully_enqueued = False
        self.__fully_produced = False
        self.__chunks_to_upload = []
        self.__n_total_chunks = 0
        self.__chunk_infos = None

    def add_chunks_to_upload(self,chunks_to_add=None,chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE) :
        """
        Add chunks from this file to the list of chunks to upload, possibly with some selection

        chunks_to_add = a list of chunk indices to add to the list to be uploaded
                        (Default=None adds all chunks)
        """
        if self.__chunk_infos is None :
            try :
                self._build_list_of_file_chunks(chunk_size)
            except Exception :
                self.logger.info(traceback.format_exc())
                fp = self.filepath.relative_to(self.__rootdir) if self.__rootdir is not None else self.filepath
                errmsg = f'ERROR: was not able to break {fp} into chunks for uploading. '
                errmsg+= 'Check log lines above for details on what went wrong. File will not be uploaded.'
                self.logger.error(errmsg)
                self.__to_upload = False
                return
        #add the chunks to the final list as DataFileChunk objects
        for ic,c in enumerate(self.__chunk_infos,start=1) :
            if chunks_to_add is None or ic in chunks_to_add :
                self.__chunks_to_upload.append(DataFileChunk(self.filepath,self.filename,self.__file_hash,
                                                            c[0],c[1],c[2],c[3],ic,self.__n_total_chunks,
                                                            rootdir=self.__rootdir,
                                                            filename_append=self.__filename_append))
        if len(self.__chunks_to_upload)>0 and self.__fully_enqueued :
            self.__fully_enqueued = False

    def enqueue_chunks_for_upload(self,queue,n_threads=None,chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE,
                                   queue_full_timeout=0.001) :
        """
        Add chunks of this file to a given upload queue. 
        If the file runs out of chunks it will be marked as fully enqueued.
        If the given queue is full this function will do absolutely nothing and will just return.

        Possible keyword arguments:
        n_threads          = the number of threads running during uploading; at most 5*this number of chunks will be 
                             added per call to this function if this argument isn't given, every chunk will be added
        chunk_size         = the size of each file chunk in bytes 
                             (used to create the list of file chunks if it doesn't already exist)
                             the default value will be used if this argument isn't given
        queue_full_timeout = amount of time to wait if the queue is full and new messages can't be added
        """
        if self.__fully_enqueued :
            warnmsg = f'WARNING: enqueue_chunks_for_upload called for fully enqueued file {self.filepath}, '
            warnmsg+= 'nothing else will be added.'
            self.logger.warning(warnmsg)
            return
        if queue.full() :
            time.sleep(queue_full_timeout)
            return
        if len(self.__chunks_to_upload)==0 :
            self.add_chunks_to_upload(chunk_size=chunk_size)
        if n_threads is not None :
            n_chunks_to_add = 5*n_threads
        else :
            n_chunks_to_add = len(self.__chunks_to_upload)
        ic = 0
        while len(self.__chunks_to_upload)>0 and ic<n_chunks_to_add :
            while queue.full() :
                time.sleep(queue_full_timeout)
            queue.put(self.__chunks_to_upload.pop(0))
            ic+=1
        if len(self.__chunks_to_upload)==0 :
            self.__fully_enqueued = True
    
    def upload_whole_file(self,config_path,topic_name,
                          n_threads=RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS,
                          chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE) :
        """
        Chunk and upload an entire file on disk to a broker's topic.

        config_path = path to the config file to use in defining the producer
        topic_name  = name of the topic to produce messages to
        n_threads   = the number of threads to run at once during uploading
        chunk_size  = the size of each file chunk in bytes
        """
        startup_msg = f"Uploading entire file {self.filepath} to {topic_name} in {chunk_size} byte chunks "
        startup_msg+=f"using {n_threads} threads...."
        self.logger.info(startup_msg)
        #add all the chunks to the upload queue
        upload_queue = Queue()
        self.enqueue_chunks_for_upload(upload_queue,chunk_size=chunk_size)
        #add "None" to the queue for each thread as the final values
        for ti in range(n_threads) :
            upload_queue.put(None)
        #produce all the messages in the queue using multiple threads
        producer_group = ProducerGroup(config_path,logger=self.logger)
        producers = []
        upload_threads = []
        for ti in range(n_threads) :
            producers.append(producer_group.get_new_producer())
            t = Thread(target=producers[-1].produce_from_queue, args=(upload_queue,topic_name))
            t.start()
            upload_threads.append(t)
        #join the threads
        for ut in upload_threads :
            ut.join()
        self.logger.info('Waiting for all enqueued messages to be delivered (this may take a moment)....')
        for producer in producers :
            producer.flush(timeout=-1) #don't leave the function until all messages have been sent/received
            producer.close()
        producer_group.close()
        self.logger.info('Done!')

    #################### PRIVATE HELPER FUNCTIONS ####################
    
    def _build_list_of_file_chunks(self,chunk_size) :
        """
        Build the list of DataFileChunks for this file

        chunk_size = the size of each chunk in bytes
        """
        #first make sure the choices of select_bytes are valid if necessary 
        #and sort them by their start byte to keep the file hash in order
        if self.select_bytes!=[] :
            if type(self.select_bytes)!=list :
                self.logger.error(f'ERROR: select_bytes={self.select_bytes} but is expected to be a list!',ValueError)
            for sbt in self.select_bytes :
                if type(sbt)!=tuple or len(sbt)!=2 :
                    errmsg = f'ERROR: found {sbt} in select_bytes but all elements are expected to be two-entry tuples!'
                    self.logger.error(errmsg,ValueError)
                elif sbt[0]>=sbt[1] :
                    errmsg = f'ERROR: found {sbt} in select_bytes but start byte cannot be >= stop byte!'
                    self.logger.error(errmsg,ValueError)
            sorted_select_bytes = sorted(self.select_bytes,key=lambda x: x[0])
        #start a hash for the file and the lists of chunks
        file_hash = sha512()
        chunks = []
        isb = 0 #index for the current sorted_select_bytes entry if necessary
        #read the binary data in the file as chunks of the given size, adding each chunk to the list 
        with open(self.filepath,'rb') as fp :
            chunk_offset = 0
            file_offset = 0 if self.select_bytes==[] else sorted_select_bytes[isb][0]
            n_bytes_to_read = chunk_size 
            if self.select_bytes!=[] :
                n_bytes_to_read = min(chunk_size,sorted_select_bytes[isb][1]-file_offset)
            chunk = fp.read(n_bytes_to_read)
            while len(chunk) > 0 :
                file_hash.update(chunk)
                chunk_hash = sha512()
                chunk_hash.update(chunk)
                chunk_hash = chunk_hash.digest()
                chunk_length = len(chunk)
                chunks.append([chunk_hash,file_offset,chunk_offset,chunk_length])
                chunk_offset += chunk_length
                file_offset += chunk_length
                if self.select_bytes!=[] and file_offset==sorted_select_bytes[isb][1] :
                    isb+=1
                    if isb>(len(sorted_select_bytes)-1) :
                        break
                    file_offset=sorted_select_bytes[isb][0]
                n_bytes_to_read = chunk_size 
                if self.select_bytes!=[] :
                    n_bytes_to_read = min(chunk_size,sorted_select_bytes[isb][1]-file_offset)
                fp.seek(file_offset)
                chunk = fp.read(n_bytes_to_read)
        file_hash = file_hash.digest()
        self.logger.info(f'File {self.filepath} has a total of {len(chunks)} chunks')
        #set the hash for the file
        self.__file_hash = file_hash
        #set the total number of chunks for this file
        self.__n_total_chunks = len(chunks)
        #build the list of all of the chunk infos for the file
        self.__chunk_infos = []
        for c in chunks :
            self.__chunk_infos.append(c)

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls) :
        superargs,superkwargs = super().get_command_line_arguments()
        args = [*superargs,'filepath','config','topic_name','chunk_size']
        kwargs = {**superkwargs,'n_threads':RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS}
        return args,kwargs

    @classmethod
    def run_from_command_line(cls,args=None) :
        """
        Run the upload data file directly from the command line
        """
        #make the argument parser
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        #make the DataFile for the single specified file
        upload_file = cls(args.filepath)
        #chunk and upload the file
        upload_file.upload_whole_file(args.config,args.topic_name,
                                      n_threads=args.n_threads,
                                      chunk_size=args.chunk_size)
        upload_file.logger.info(f'Done uploading {args.filepath}')


#################### MAIN METHOD TO RUN FROM COMMAND LINE ####################

def main(args=None) :
    UploadDataFile.run_from_command_line(args)

if __name__=='__main__' :
    main()



#imports
import sys, os, pathlib, shutil, ctypes.util
from subprocess import CalledProcessError
from .config import SERVICE_CONST
from .utilities import run_cmd_in_subprocess
from .service_manager_base import ServiceManagerBase

class WindowsServiceManager(ServiceManagerBase) :
    """
    Class for working with Windows Services
    """

    @property
    def env_var_names(self):
        for evn in super().env_var_names :
            yield evn
        #get the names of environment variables from the env_var file
        if self.env_var_filepath.is_file() :
            with open(self.env_var_filepath,'r') as fp :
                lines = fp.readlines()
            for line in lines :
                try :
                    linestrip = line.strip()
                    if linestrip!='' :
                        yield linestrip
                except :
                    pass

    def install_service(self) :
        super().install_service()
        #if it doesn't exist there yet, copy the libsodium.dll file to C:\Windows\system32
        self.__copy_libsodium_dll_to_system32()
        #find or install NSSM to run the executable
        self.__find_install_NSSM()
        #run NSSM to install the service
        cmd = f'{SERVICE_CONST.NSSM_PATH} install {self.service_name} \"{sys.executable}\" \"{self.exec_fp}\"'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        run_cmd_in_subprocess(['powershell.exe',
                               f'{SERVICE_CONST.NSSM_PATH} set {self.service_name} DisplayName {self.service_name}'],
                               logger=self.logger)
        self.logger.info(f'Done installing {self.service_name}')

    def start_service(self) :
        self.logger.info(f'Starting {self.service_name}...')
        #start the service using net
        cmd = f'net start {self.service_name}'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info(f'Done starting {self.service_name}')

    def service_status(self) :
        #find or install NSSM in the current directory
        self.__find_install_NSSM()
        #get the service status
        cmd = f'{SERVICE_CONST.NSSM_PATH} status {self.service_name}'
        result = run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info(f'{self.service_name} status: {result.decode()}')

    def stop_service(self) :
        self.logger.info(f'Stopping {self.service_name}...')
        #stop the service using net
        cmd = f'net stop {self.service_name}'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info(f'Done stopping {self.service_name}')

    def remove_service(self,remove_env_vars=False,remove_install_args=False,remove_nssm=False) :
        self.logger.info(f'Removing {self.service_name}...')
        #find or install NSSM in the current directory
        self.__find_install_NSSM()
        #using NSSM
        cmd = f'{SERVICE_CONST.NSSM_PATH} remove {self.service_name} confirm'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info('Service removed')
        #remove NSSM if requested
        if remove_nssm :
            if SERVICE_CONST.NSSM_PATH.is_file() :
                try :
                    run_cmd_in_subprocess(['powershell.exe',f'del {SERVICE_CONST.NSSM_PATH}'],logger=self.logger)
                except CalledProcessError :
                    msg = f'WARNING: failed to delete {SERVICE_CONST.NSSM_PATH}. '
                    msg+= 'You are free to delete it manually if you would like.'
                    self.logger.info(msg)
            else :
                msg = f'NSSM does not exist at {SERVICE_CONST.NSSM_PATH} so it will not be removed'
                self.logger.info(msg)
        else :
            if SERVICE_CONST.NSSM_PATH.is_file() :
                msg = f'NSSM executable at {SERVICE_CONST.NSSM_PATH} will be retained'
                self.logger.info(msg)
        super().remove_service(remove_env_vars=remove_env_vars,remove_install_args=remove_install_args)

    def _write_env_var_file(self) :
        code = ''
        for evn in self.env_var_names :
            val = os.path.expandvars(f'${evn}')
            if val==f'${evn}' :
                raise RuntimeError(f'ERROR: value not found for expected environment variable {evn}!')
            code += f'{evn}\n'
        if code=='' :
            return False
        with open(self.env_var_filepath,'w') as fp :
            fp.write(code)
        return True

    def __copy_libsodium_dll_to_system32(self) :
        """
        Ensure that the libsodium.dll file exists in C:\Windows\system32
        (Needed to properly load it when running as a service)
        """
        system32_path = pathlib.Path(r'C:\Windows\system32')/'libsodium.dll'
        if system32_path.is_file() :
            return
        current_env_dll = ctypes.util.find_library('libsodium')
        if current_env_dll is not None :
            shutil.copy(pathlib.Path(current_env_dll),system32_path)
        else :
            errmsg = f'ERROR: could not locate libsodium DLL to copy to system32 folder for {self.service_name}!'
            self.logger.error(errmsg,FileNotFoundError)

    def __find_install_NSSM(self,move_local=True) :
        """
        Ensure the NSSM executable exists in the expected location.
        If it exists in the current directory, move it to the expected location.
        If it doesn't exist anywhere, try to download it from the web, but that's finicky in powershell.
        """
        if SERVICE_CONST.NSSM_PATH.is_file() :
            return
        else :
            if (pathlib.Path()/SERVICE_CONST.NSSM_PATH.name).is_file() and move_local :
                (pathlib.Path()/SERVICE_CONST.NSSM_PATH.name).replace(SERVICE_CONST.NSSM_PATH)
                return self.__find_install_NSSM(move_local=False)
            SERVICE_CONST.LOGGER.info(f'Installing NSSM from {SERVICE_CONST.NSSM_DOWNLOAD_URL}...')
            nssm_zip_file_name = SERVICE_CONST.NSSM_DOWNLOAD_URL.split('/')[-1]
            run_cmd_in_subprocess(['powershell.exe',
                                   '[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Tls12'],
                                   logger=self.logger)
            cmd_tuples = [
                (f'curl {SERVICE_CONST.NSSM_DOWNLOAD_URL} -O',
                f'Invoke-WebRequest -Uri {SERVICE_CONST.NSSM_DOWNLOAD_URL} -OutFile {nssm_zip_file_name}'),
                (f'tar -xf {pathlib.Path() / nssm_zip_file_name}',
                f'Expand-Archive {nssm_zip_file_name} -DestinationPath {pathlib.Path().resolve()}'),
                (f'del {nssm_zip_file_name}',
                f'Remove-Item -Path {nssm_zip_file_name}'),
                (f'move {pathlib.Path() / nssm_zip_file_name.rstrip(".zip") / "win64" / "nssm.exe"} {pathlib.Path()}',
                f'''Move-Item -Path {pathlib.Path()/nssm_zip_file_name.rstrip(".zip")/"win64"/"nssm.exe"} \
                    -Destination {SERVICE_CONST.NSSM_PATH}'''),
                (f'rmdir /S /Q {nssm_zip_file_name.rstrip(".zip")}',
                f'Remove-Item -Recurse -Force {nssm_zip_file_name.rstrip(".zip")}'),
            ]
            for cmd in cmd_tuples :
                try :
                    run_cmd_in_subprocess(['powershell.exe',cmd[1]],logger=self.logger)
                except CalledProcessError :
                    run_cmd_in_subprocess(cmd[0],shell=True,logger=self.logger)
            SERVICE_CONST.LOGGER.debug('Done installing NSSM')



#imports
from abc import ABC, abstractmethod

class Producible(ABC) :
    """
    Small utility class for anything that can be Produced as a message to a topic
    """

    @property
    @abstractmethod
    def msg_key(self) :
        """
        The key of the object when it's represented as a message
        This can be something that needs to be serialized still
        Not implemented in base class
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def msg_value(self) :
        """
        The value of the object when it's represented as a message
        This can be something that needs to be serialized still
        Not implemented in base class
        """
        raise NotImplementedError

    @property
    def callback_kwargs(self) :
        """
        A dictionary of keyword arguments that should be sent to the callback function 
        for Producers producing messages of this Producible
        Empty for the base class
        """
        return {}

    @abstractmethod
    def get_log_msg(self,print_every=None) :
        """
        Given some (optional) "print_every" variable, return the string that should be logged for this message
        If "None" is returned nothing will be logged.

        Not implemented in base class
        """
        return None


#imports
import pathlib

class RunConstants :

    @property
    def CONFIG_FILE_EXT(self) :
        return '.config'

    @property
    def CONFIG_FILE_DIR(self) :
        return pathlib.Path(__file__).parent.parent / 'kafka_wrapper' / 'config_files'

    @property
    def DEFAULT_N_THREADS(self) :
        return 2

    @property
    def DEFAULT_UPDATE_SECONDS(self) :
        #how many seconds to wait by default between printing the "still alive" character/message for a running process
        return 300     

RUN_CONST = RunConstants()



#imports
import sys, os, pathlib, textwrap
from .config import SERVICE_CONST
from .utilities import run_cmd_in_subprocess
from .service_manager_base import ServiceManagerBase

class LinuxServiceManager(ServiceManagerBase) :
    """
    Class for working with Linux Services/daemons
    """

    @property
    def env_var_names(self):
        for evn in super().env_var_names :
            yield evn
        #get the names of environment variables from the env_var file
        if self.env_var_filepath.is_file() :
            with open(self.env_var_filepath,'r') as fp :
                lines = fp.readlines()
            for line in lines :
                try :
                    linesplit = (line.strip()).split('=')
                    if len(linesplit)==2 :
                        yield linesplit[0]
                except :
                    pass

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.daemon_working_dir_filepath = SERVICE_CONST.WORKING_DIR/f'{self.service_name}.service'
        self.daemon_filepath = SERVICE_CONST.DAEMON_SERVICE_DIR/self.daemon_working_dir_filepath.name

    def install_service(self) :
        super().install_service()
        #make sure systemd is running
        self.__check_systemd_installed()
        #write the daemon file pointing to the executable
        self.__write_daemon_file()
        #enable the service
        run_cmd_in_subprocess(['sudo','systemctl','daemon-reload'],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','enable',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'Done installing {self.service_name}')

    def start_service(self) :
        self.logger.info(f'Starting {self.service_name}...')
        self.__check_systemd_installed()
        run_cmd_in_subprocess(['sudo','systemctl','daemon-reload'],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','start',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'Done starting {self.service_name}')

    def service_status(self) :
        self.__check_systemd_installed()
        result = run_cmd_in_subprocess(['sudo','systemctl','status',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'{self.service_name} status: {result.decode()}')

    def stop_service(self) :
        self.logger.info(f'Stopping {self.service_name}...')
        self.__check_systemd_installed()
        run_cmd_in_subprocess(['sudo','systemctl','stop',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'Done stopping {self.service_name}')

    def remove_service(self,remove_env_vars=False,remove_install_args=False,remove_nssm=False) :
        self.logger.info(f'Removing {self.service_name}...')
        self.__check_systemd_installed()
        run_cmd_in_subprocess(['sudo','systemctl','disable',f'{self.service_name}.service'],logger=self.logger)
        if self.daemon_filepath.exists() :
            run_cmd_in_subprocess(['sudo','rm','-f',str(self.daemon_filepath)],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','daemon-reload'],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','reset-failed'],logger=self.logger)
        self.logger.info('Service removed')
        if remove_nssm :
            warnmsg = "WARNING: requested to remove NSSM along with the Service, "
            warnmsg+= "but Linux Services don't install NSSM so nothing was done."
            self.logger.warning(warnmsg)
        super().remove_service(remove_env_vars=remove_env_vars,remove_install_args=remove_install_args)

    def _write_env_var_file(self) :
        code = ''
        for evn in self.env_var_names :
            val = os.path.expandvars(f'${evn}')
            if val==f'${evn}' :
                raise RuntimeError(f'ERROR: value not found for expected environment variable {evn}!')
            code += f'{evn}={val}\n'
        if code=='' :
            return False
        with open(self.env_var_filepath,'w') as fp :
            fp.write(code)
        run_cmd_in_subprocess(['chmod','go-rwx',self.env_var_filepath])
        return True 

    def __check_systemd_installed(self) :
        """
        Raises an error if systemd is not installed (systemd is needed to control Linux daemons)
        """
        check = run_cmd_in_subprocess(['ps','--no-headers','-o','comm','1'],logger=self.logger)
        if check.decode().rstrip()!='systemd' :
            errmsg = 'ERROR: Installing programs as Services ("daemons") on Linux requires systemd!'
            errmsg = 'You can install systemd with "sudo apt install systemd" (or similar) and try again.'
            self.logger.error(errmsg,RuntimeError)

    def __write_daemon_file(self) :
        """
        Write the Unit/.service file to the daemon directory that calls the Python executable
        """
        #make sure the directory to hold the file exists
        if not SERVICE_CONST.DAEMON_SERVICE_DIR.is_dir() :
            SERVICE_CONST.LOGGER.info(f'Creating a new daemon service directory at {SERVICE_CONST.DAEMON_SERVICE_DIR}')
            SERVICE_CONST.DAEMON_SERVICE_DIR.mkdir(parents=True)
        #write out the file pointing to the python executable
        code = f'''\
            [Unit]
            Description = {self.service_dict['class'].__doc__.strip()}
            Requires = network-online.target remote-fs.target
            After = network-online.target remote-fs.target

            [Service]
            Type = simple
            User = {os.path.expandvars('$USER')}
            ExecStart = {sys.executable} {self.exec_fp}'''
        if self.env_vars_needed :
            code+=f'''\n\
            EnvironmentFile = {self.env_var_filepath}'''
        code+=f'''\n\
            WorkingDirectory = {pathlib.Path().resolve()}
            Restart = on-failure
            RestartSec = 30

            [Install]
            WantedBy = multi-user.target'''
        with open(self.daemon_working_dir_filepath,'w') as fp :
            fp.write(textwrap.dedent(code))
        run_cmd_in_subprocess(['sudo','mv',str(self.daemon_working_dir_filepath),str(self.daemon_filepath.parent)],
                              logger=self.logger)



#imports
import pathlib, importlib, pkg_resources
from inspect import isclass
from ..utilities.logging import Logger

class ServicesConstants :
    """
    Constants for working with services
    """

    def __init__(self) :
        #make the Service dictionaries to use
        self.service_dicts = []
        for script in pkg_resources.iter_entry_points('console_scripts') :
            if script.dist.key == 'openmsistream' :
                if script.name in ('InstallService','ManageService','ProvisionNode') :
                    continue
                scriptstr = str(script)
                cmd = (scriptstr.split())[0]
                path = ((scriptstr.split())[2].split(':'))[0]
                funcname = (((scriptstr.split())[2]).split(':'))[1]
                module = importlib.import_module(path)
                run_classes = [getattr(module,x) for x in dir(module) 
                               if isclass(getattr(module,x)) and getattr(module,x).__name__==script.name]
                if len(run_classes)!=1 :
                    errmsg = f'ERROR: could not determine class for script {cmd} in file {path}! '
                    errmsg+= f'Possibilities found: {run_classes}'
                    raise RuntimeError(errmsg)
                self.service_dicts.append({'script_name':cmd,
                                           'class':run_classes[0],
                                           'filepath':path,
                                           'func_name':funcname})
        #make the logger to use
        self.logger = Logger('Services',logger_filepath=self.WORKING_DIR/'Services.log')

    @property
    def WORKING_DIR(self) :
        return (pathlib.Path(__file__).parent/'working_dir').resolve()
    @property
    def AVAILABLE_SERVICES(self) :
        return self.service_dicts # A dictionary with details of the services that are available
    @property
    def LOGGER(self) :
        return self.logger # A shared logger object to use with a constant file
    @property
    def NSSM_DOWNLOAD_URL(self) :
        return 'https://nssm.cc/release/nssm-2.24.zip' # The URL to use for downloading NSSM when needed
    @property
    def NSSM_PATH(self) :
        return (self.WORKING_DIR / 'nssm.exe')
    @property
    def ERROR_LOG_STEM(self) :
        return '_ERROR_LOG.txt'
    @property
    def SERVICE_EXECUTABLE_NAME_STEM(self) :
        return '_service_executable.py'
    @property
    def DAEMON_SERVICE_DIR(self) :
        return pathlib.Path('/etc/systemd/system/')

SERVICE_CONST = ServicesConstants()


#imports
import os
from hashlib import sha512
from contextlib import nullcontext
from abc import ABC, abstractmethod
from .config import DATA_FILE_HANDLING_CONST
from .data_file import DataFile

class DownloadDataFile(DataFile,ABC) :
    """
    Class to represent a data file that will be read as messages from a topic
    """

    #################### PROPERTIES AND STATIC METHODS ####################

    @staticmethod
    def get_full_filepath(dfc) :
        """
        Return the full filepath of a file that will be written to disk given one of its DataFileChunks
        """
        if dfc.filename_append=='' :
            return dfc.filepath 
        else :
            filename_split = dfc.filepath.name.split('.')
            full_fp = dfc.filepath.parent/(filename_split[0]+dfc.filename_append+'.'+('.'.join(filename_split[1:])))
            return full_fp

    @property
    def full_filepath(self) :
        return self.__full_filepath

    @property
    def subdir_str(self) :
        return self.__subdir_str

    @property
    @abstractmethod
    def check_file_hash(self) :
        pass #the hash of the data in the file after it was read; not implemented in the base class

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        #start an empty set of this file's downloaded offsets
        self._chunk_offsets_downloaded = []
        self.__full_filepath = None
        self.__subdir_str = None

    def add_chunk(self,dfc,thread_lock=nullcontext(),*args,**kwargs) :
        """
        A function to process a chunk that's been read from a topic
        Returns a number of codes based on what effect adding the chunk had
        
        This function calls _on_add_chunk, 
        
        dfc = the DataFileChunk object whose data should be added
        thread_lock = the lock object to acquire/release so that race conditions don't affect 
                      reconstruction of the files (optional, only needed if running this function asynchronously)
        """
        #if this chunk's offset has already been written to disk, return the "already written" code
        with thread_lock :
            already_written = dfc.chunk_offset_write in self._chunk_offsets_downloaded
        if already_written :
            return DATA_FILE_HANDLING_CONST.CHUNK_ALREADY_WRITTEN_CODE
        #the filepath of this DownloadDataFile and of the given DataFileChunk must match
        if dfc.filepath!=self.filepath :
            errmsg = f'ERROR: filepath mismatch between data file chunk with {dfc.filepath} and '
            errmsg+= f'data file with {self.filepath}'
            self.logger.error(errmsg,ValueError)
        #modify the filepath to include any append to the name
        full_filepath = self.__class__.get_full_filepath(dfc)
        if self.__full_filepath is None :
            self.__full_filepath = full_filepath
            self.filename = self.__full_filepath.name
        elif self.__full_filepath!=full_filepath :
            errmsg = f'ERROR: filepath for data file chunk {dfc.chunk_i}/{dfc.n_total_chunks} with offset '
            errmsg+= f'{dfc.chunk_offset_write} is {full_filepath} but the file being reconstructed is '
            errmsg+= f'expected to have filepath {self.__full_filepath}'
            self.logger.error(errmsg,ValueError)
        #add the subdirectory string to this file
        if self.__subdir_str is None :
            self.__subdir_str = dfc.subdir_str
        elif self.__subdir_str!=dfc.subdir_str :
            errmsg = f"Mismatched subdirectory strings! From file = {self.__subdir_str}, from chunk = {dfc.subdir_str}"
            self.logger.error(errmsg,ValueError)
        #acquire the thread lock to make sure this process is the only one dealing with this particular file
        with thread_lock:
            #call the function to actually add the chunk
            self._on_add_chunk(dfc,*args,**kwargs)
            #add the offset of the added chunk to the set of reconstructed file chunks
            self._chunk_offsets_downloaded.append(dfc.chunk_offset_write)
            last_chunk = len(self._chunk_offsets_downloaded)==dfc.n_total_chunks
        #if this chunk was the last that needed to be added, check the hashes
        if last_chunk :
            if self.check_file_hash!=dfc.file_hash :
                return DATA_FILE_HANDLING_CONST.FILE_HASH_MISMATCH_CODE
            else :
                return DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE
        else :
            return DATA_FILE_HANDLING_CONST.FILE_IN_PROGRESS

    #################### PRIVATE HELPER FUNCTIONS ####################

    @abstractmethod
    def _on_add_chunk(dfc,*args,**kwargs) :
        """
        A function to actually process a new chunk being added to the file
        This function is executed while a thread lock is acquired so it will never run asynchronously
        Also any DataFileChunks passed to this function are guaranteed to have unique offsets
        Not implemented in the base class
        """
        pass

class DownloadDataFileToDisk(DownloadDataFile) :
    """
    Class to represent a data file that will be reconstructed on disk using messages read from a topic
    """

    #################### PROPERTIES ####################

    @property
    def check_file_hash(self) :
        check_file_hash = sha512()
        with open(self.full_filepath,'rb') as fp :
            data = fp.read()
        check_file_hash.update(data)
        return check_file_hash.digest()

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        #create the parent directory of the file if it doesn't exist yet (in case the file is in a new subdirectory)
        if not self.filepath.parent.is_dir() :
            self.filepath.parent.mkdir(parents=True)

    def _on_add_chunk(self,dfc) :
        """
        Add the data from a given file chunk to this file on disk
        """
        mode = 'r+b' if self.full_filepath.is_file() else 'w+b'
        with open(self.full_filepath,mode) as fp :
            fp.seek(dfc.chunk_offset_write)
            fp.write(dfc.data)
            fp.flush()
            os.fsync(fp.fileno())
            fp.close()

class DownloadDataFileToMemory(DownloadDataFile) :
    """
    Class to represent a data file that will be held in memory and populated by the contents of messages from a topic
    """

    #################### PROPERTIES ####################

    @property
    def bytestring(self) :
        if self.__bytestring is None :
            self.__create_bytestring()
        return self.__bytestring

    @property
    def check_file_hash(self) :
        check_file_hash = sha512()
        check_file_hash.update(self.bytestring)
        return check_file_hash.digest()

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        #start a dictionary of the file data by their offsets
        self.__chunk_data_by_offset = {}
        #placeholder for the eventual full data bytestring
        self.__bytestring = None

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _on_add_chunk(self,dfc) :
        """
        Add the data from a given file chunk to the dictionary of data by offset
        """
        self.__chunk_data_by_offset[dfc.chunk_offset_write] = dfc.data

    def __create_bytestring(self) :
        """
        Makes all of the data held in the dictionary into a single bytestring ordered by offset of each chunk
        """
        bytestring = b''
        for data in [self.__chunk_data_by_offset[offset] for offset in sorted(self.__chunk_data_by_offset.keys())] :
            bytestring+=data
        self.__bytestring = bytestring



#imports
from ..utilities.logging import LogOwner
from .openmsistream_consumer import OpenMSIStreamConsumer

class ConsumerGroup(LogOwner) :
    """
    Class for working with a group of consumers
    """

    @property
    def topic_name(self) :
        return self.__topic_name

    def __init__(self,config_path,topic_name,*args,consumer_group_ID='create_new',**kwargs) :
        """
        arguments:
        config_path = path to the config file that should be used to define the consumer group
        topic_name  = name of the topic to consume messages from

        keyword arguments:
        consumer_group_ID = ID to use for all consumers in the group (a new & unique ID is created by default)
        """
        super().__init__(*args,**kwargs)
        self.__topic_name = topic_name
        self.__c_args, self.__c_kwargs = OpenMSIStreamConsumer.get_consumer_args_kwargs(config_path,
                                                                                        group_id=consumer_group_ID,
                                                                                        logger=self.logger)

    def get_new_subscribed_consumer(self) :
        """
        Return a new Consumer, subscribed to the topic and with the shared group ID
        Call this function from a child thread to get thread-independent Consumers 
        This function just creates and subscribes the Consumer. Polling it, closing 
        it, and everything else must be handled by whatever calls this function.
        """
        consumer = OpenMSIStreamConsumer(*self.__c_args,**self.__c_kwargs)
        consumer.subscribe([self.__topic_name])
        return consumer

    def close(self) :
        try :
            self.__c_kwargs['kafkacrypto'].close()
        except :
            pass
        finally :
            self.__c_kwargs['kafkacrypto'] = None



#imports
import time
from confluent_kafka import SerializingProducer
from kafkacrypto import KafkaProducer
from ..utilities.logging import LogOwner
from .utilities import add_kwargs_to_configs, default_producer_callback, make_callback
from .producible import Producible
from .config_file_parser import KafkaConfigFileParser
from .openmsistream_kafka_crypto import OpenMSIStreamKafkaCrypto
from .serialization import CompoundSerializer

class OpenMSIStreamProducer(LogOwner) :
    """
    Convenience class for working with a Producer of some type
    """

    POLL_EVERY = 5 # poll the producer at least every 5 calls to produce

    def __init__(self,producer_type,configs,kafkacrypto=None,**kwargs) :
        """
        producer_type = the type of Producer underlying this object
        configs = a dictionary of configurations to pass to the producer to create it
        """
        super().__init__(**kwargs)
        if producer_type==KafkaProducer :
            if kafkacrypto is None :
                self.logger.error('ERROR: creating a KafkaProducer requires holding onto its KafkaCrypto objects!')
            self.__kafkacrypto = kafkacrypto
            self.__producer = producer_type(**configs)
        elif producer_type==SerializingProducer :
            self.__producer = producer_type(configs)
        else :
            errmsg=f'ERROR: Unrecognized producer type {producer_type} for OpenMSIStreamProducer!'
            self.logger.error(errmsg,ValueError)
        # poll the producer at least every 5 calls to produce
        self.__poll_counter = 0

    @staticmethod
    def get_producer_args_kwargs(config_file_path,logger=None,**kwargs) :
        """
        config_file_path = path to the config file to use in defining this producer

        any keyword arguments will be added to the final producer configs (with underscores replaced with dots)
        """
        parser = KafkaConfigFileParser(config_file_path,logger=logger)
        ret_kwargs = {}
        #get the broker and producer configurations
        all_configs = {**parser.broker_configs,**parser.producer_configs}
        all_configs = add_kwargs_to_configs(all_configs,logger,**kwargs)
        #all_configs['debug']='broker,topic,msg'
        #if there are configs for KafkaCrypto, use a KafkaProducer
        if parser.kc_config_file_str is not None :
            if logger is not None :
                logger.debug(f'Produced messages will be encrypted using configs at {parser.kc_config_file_str}')
            kc = OpenMSIStreamKafkaCrypto(parser.broker_configs,parser.kc_config_file_str)
            if 'key.serializer' in all_configs.keys() :
                keyser = CompoundSerializer(all_configs.pop('key.serializer'),kc.key_serializer)
            else :
                keyser = kc.key_serializer
            if 'value.serializer' in all_configs.keys() :
                valser = CompoundSerializer(all_configs.pop('value.serializer'),kc.value_serializer)
            else :
                valser = kc.value_serializer
            all_configs['key_serializer']=keyser
            all_configs['value_serializer']=valser
            ret_args = [KafkaProducer,all_configs]
            ret_kwargs['kafkacrypto']=kc
        #otherwise use a SerializingProducer
        else :
            ret_args = [SerializingProducer,all_configs]
        ret_kwargs['logger'] = logger
        return ret_args, ret_kwargs

    @classmethod
    def from_file(cls,*args,**kwargs) :
        args_to_use, kwargs_to_use = OpenMSIStreamProducer.get_producer_args_kwargs(*args,**kwargs)
        return cls(*args_to_use,**kwargs_to_use)

    def produce_from_queue(self,queue,topic_name,callback=None,print_every=1000,timeout=60,retry_sleep=5) :
        """
        Get Producible objects from a given queue and Produce them to the given topic.
        Runs until "None" is pulled from the Queue
        Meant to be run in multiple threads in parallel

        queue       = the Queue holding objects that should be Produced
        topic_name  = the name of the topic to Produce to
        callback    = a function that should be called for each message upon recognition by the broker
        print_every = how often to print/log progress messages
        timeout     = max time (s) to wait for the message to be produced in the event of (repeated) BufferError(s)
        retry_sleep = how long (s) to wait between produce attempts if one fails with a BufferError
        """
        #get the next object from the Queue
        obj = queue.get()
        #loop until "None" is pulled from the Queue
        while obj is not None :
            if isinstance(obj,Producible) :
                #log a line about this message if applicable
                logmsg = obj.get_log_msg(print_every)
                if logmsg is not None :
                    self.logger.info(logmsg)
                #get the Producible's callback arguments and set the callback to use
                if callback is None :
                    callback_to_use = make_callback(default_producer_callback,logger=self.logger,**obj.callback_kwargs)
                else :
                    callback_to_use = make_callback(callback,**obj.callback_kwargs)
                #produce the message to the topic
                success=False; total_wait_secs=0 
                while (not success) and total_wait_secs<timeout :
                    try :
                        self.produce(topic=topic_name,key=obj.msg_key,value=obj.msg_value,on_delivery=callback_to_use)
                        success=True
                    except BufferError :
                        n_new_callbacks = self.poll(0)
                        time.sleep(retry_sleep)
                        if n_new_callbacks is None or n_new_callbacks==0 :
                            total_wait_secs+=retry_sleep
                        else :
                            total_wait_secs = 0
                if not success :
                    warnmsg = f'WARNING: message with key {obj.msg_key} failed to buffer for more than '
                    warnmsg+= f'{total_wait_secs}s with no new callbacks served. This message will be re-enqueued.'
                    self.logger.warning(warnmsg)
                    queue.put(obj)
                self.__poll_counter+=1
                if self.__poll_counter%self.POLL_EVERY==0 :
                    n_new_callbacks = self.poll(0)
                    self.__poll_counter = 0
            else :
                warnmsg = f'WARNING: found an object of type {type(obj)} in a Producer queue that should only contain '
                warnmsg+= 'Producible objects. This object will be skipped!'
                self.logger.warning(warnmsg)
            #get the next object in the Queue
            obj = queue.get()
        queue.task_done()

    def produce(self,*args,topic,key,value,**kwargs) :
        if isinstance(self.__producer,KafkaProducer) :
            key = self.__producer.ks(topic,key)
            value = self.__producer.vs(topic,value)
        return self.__producer.produce(*args,topic=topic,key=key,value=value,**kwargs)
    
    def poll(self,*args,**kwargs) :
        return self.__producer.poll(*args,**kwargs)
    
    def flush(self,*args,**kwargs) :
        return self.__producer.flush(*args,**kwargs)

    def close(self) :
        try :
            if self.__kafkacrypto :
                self.__kafkacrypto.close()
        except :
            pass
        finally :
            self.__kafkacrypto = None



#imports
import setuptools, site

version = '0.9.1.8'

site.ENABLE_USER_SITE = True #https://www.scivision.dev/python-pip-devel-user-install/

long_description = ''
with open('README.md', 'r') as readme :
    for il,line in enumerate(readme.readlines(),start=1) :
        if il>=12 :
            long_description+=line

setupkwargs = dict(
    name='openmsistream',
    packages=setuptools.find_packages(include=['openmsistream*']),
    include_package_data=True,
    version=version,
    description='Python applications for materials data streaming using Apache Kafka. Developed for Open MSI (NSF DMREF award #1921959)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Maggie Eminizer, Amir Sharifzadeh, Sam Tabrisky, Alakarthika Ulaganathan, David Elbert',
    author_email='margaret.eminizer@gmail.com',
    url='https://github.com/openmsi/openmsistream',
    download_url=f'https://github.com/openmsi/openmsistream/archive/refs/tags/v{version}.tar.gz',
    license='GNU GPLv3',
    entry_points = {
        'console_scripts' : ['UploadDataFile=openmsistream.data_file_io.upload_data_file:main',
                             'DataFileUploadDirectory=openmsistream.data_file_io.data_file_upload_directory:main',
                             'DataFileDownloadDirectory=openmsistream.data_file_io.data_file_download_directory:main',
                             'InstallService=openmsistream.services.install_service:main',
                             'ManageService=openmsistream.services.manage_service:main',
                             'ProvisionNode=openmsistream.utilities.provision_wrapper:main',
                             'S3TransferStreamProcessor=openmsistream.s3_buckets.s3_transfer_stream_processor:main',
                            ],
    },
    python_requires='>=3.7,<3.10',
    install_requires=['atomicwrites>=1.4.0',
                      'boto3>=1.23.0',
                      'confluent-kafka>=1.8.2',
                      'kafkacrypto>=0.9.9.15',
                      'methodtools',
                      'msgpack',
                     ],
    extras_require = {'test': ['beautifulsoup4',
                               'gitpython',
                               'lxml',
                               'marko[toc]',
                               'pyflakes>=2.2.0',
                               ],
                      'docs': ['sphinx>=5.0.2',
                               'sphinx_rtd_theme>=1.0.0',
                                ],
                      'dev': ['twine',
                                ],
                        },
    keywords=['data_streaming','stream_processing','apache_kafka','materials','data_science'],
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
)

setupkwargs["extras_require"]["all"] = sum(setupkwargs["extras_require"].values(), [])

setuptools.setup(**setupkwargs)



#imports
from threading import Thread

class ExceptionTrackingThread(Thread) :
    """
    A small utility class to keep track of any exceptions thrown in a child thread
    and raise them at some point
    """

    @property
    def caught_exception(self) :
        return self.__exc

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.__exc = None
    
    def run(self,*args,**kwargs) :
        try :
            super().run(*args,**kwargs)
        except Exception as e :
            self.__exc = e
    
    def join(self,*args,**kwargs) :
        super().join(*args,**kwargs)
        if self.__exc is not None :
            raise self.__exc


import os
if os.name=='nt' :
    try :
        import confluent_kafka
    except Exception :
        import traceback
        try :
            import sys, pathlib
            dll_dir = pathlib.Path(sys.executable).parent/'Lib'/'site-packages'/'confluent_kafka.libs'
            if not dll_dir.is_dir() :
                raise FileNotFoundError(f'ERROR: expected DLL directory {dll_dir} does not exist!')
            from ctypes import CDLL
            fps = []
            for fp in dll_dir.glob('*.dll') :
                if fp.name.startswith('librdkafka') :
                    fps.append(fp)
                    CDLL(str(fp))
            if len(fps)<1 :
                raise RuntimeError(f'ERROR: {dll_dir} does not contain any librdkafka DLL files to preload!')
        except Exception :
            print(f'Failed to preload librdkafka DLLs. Exception: {traceback.format_exc()}')
        try :
            import confluent_kafka
        except Exception :
            errmsg = 'Preloading librdkafka DLLs ('
            for fp in fps :
                errmsg+=f'{fp}, '
            errmsg = f'{errmsg[:-2]}) did not allow confluent_kafka to be imported! Exception: '
            errmsg+= f'{traceback.format_exc()}'
            raise ImportError(errmsg)
    _ = confluent_kafka.Producer #appease pyflakes



#imports
import os, sys, inspect, time, contextlib

def add_user_input(input_queue) :
    """
    Listen for and add user input to a queue at one second intervals
    """
    while True :
        time.sleep(1)
        input_queue.put((sys.stdin.read(1)).strip())

def populated_kwargs(given_kwargs,defaults,logger=None) :
    """
    Return a kwargs dictionary where every possible entry from the defaults has a valid value
    Can use this to make sure certain entries are present in kwargs
    """
    #for all the needed keys
    for key in defaults.keys() :
        #if it was given and is not None
        if key in given_kwargs.keys() and given_kwargs[key] is not None :
            #if there are several options for what the argument can be, check them all
            if isinstance(defaults[key],tuple) :
                #if there is a default and at least one specified class or type that the argument must be
                if len(defaults[key])>1 :
                    #if the default argument is a class
                    if inspect.isclass((defaults[key])[0]) :
                        #make sure the given argument is also a class
                        if (not inspect.isclass(given_kwargs[key])) :
                            errmsg = f'ERROR: Argument "{key}" expected to be a class, not an object!'
                            if logger is not None :
                                logger.error(errmsg,RuntimeError)
                            else :
                                raise RuntimeError(errmsg)
                        #make sure the given argument is a subtype of any of the classes specified in the defaults
                        isgood=False
                        for poss_class in (defaults[key])[1:] :
                            if issubclass(given_kwargs[key],poss_class) :
                                isgood=True
                                break
                        if not isgood :
                            errmsg = f'ERROR: Class type mismatch for argument "{key}", got {given_kwargs[key]} '
                            errmsg+= f'but expected one of {(defaults[key])[1:]}'
                            if logger is not None :
                                logger.error(errmsg,RuntimeError)
                            else :
                                raise RuntimeError(errmsg)
                    #if the default is an object, make sure the that type of the given argument matches 
                    #one of the possible types in the defaults
                    elif type(given_kwargs[key]) not in (defaults[key])[1:] :
                        errmsg = f'ERROR: Type mismatch replacing argument "{key}" with {given_kwargs[key]} '
                        errmsg+= '(expected one of '
                        for t in (defaults[key])[1:] :
                            errmsg+=f'{t}, '
                        errmsg+=f'but got {type(given_kwargs[key])})'
                        if logger is not None :
                            logger.error(errmsg,TypeError)
                        else :
                            raise TypeError(errmsg)
                #otherwise if there's just the default option it doesn't need to be a tuple
                else :
                    defaults[key] = (defaults[key])[0]
            #if the single default option is a class
            elif inspect.isclass(defaults[key]) :
                #make sure the given argument is also a class
                if (not inspect.isclass(given_kwargs[key])) :
                    errmsg = f'ERROR: Argument "{key}" expected to be a class, not an object!'
                    if logger is not None :
                        logger.error(errmsg,RuntimeError)
                    else :
                        raise RuntimeError(errmsg)
                #make sure the given argument is a subtype of the default class
                if not issubclass(given_kwargs[key],defaults[key]) :
                    errmsg = f'ERROR: Class type mismatch for argument "{key}", got {given_kwargs[key]} '
                    errmsg+= f'but expected {defaults[key]}'
                    if logger is not None :
                        logger.error(errmsg,RuntimeError)
                    else :
                        raise RuntimeError(errmsg)
            #if the single default option is just an object, make sure the argument is of the same type as the default
            elif defaults[key] is not None and type(defaults[key]) is not type(given_kwargs[key]) :
                errmsg = f'ERROR: Type mismatch replacing argument "{key}" with {given_kwargs[key]} '
                errmsg+= f'(expected type {type(defaults[key])})'
                if logger is not None :
                    logger.error(errmsg,TypeError)
                else :
                    raise TypeError(errmsg)
        #if it wasn't given, then just add it to the dictionary as the default
        else :
            if isinstance(defaults[key],tuple) :
                given_kwargs[key] = (defaults[key])[0]
            else :
                given_kwargs[key] = defaults[key]
    return given_kwargs

@contextlib.contextmanager
def cd(dir):
    """
    Change the current working directory to a different directory,
    and go back when leaving the context manager.
    """
    cdminus = os.getcwd()
    try:
        yield os.chdir(dir)
    finally:
        os.chdir(cdminus)



#imports
import uuid
from confluent_kafka import DeserializingConsumer, Message
from kafkacrypto import KafkaConsumer
from ..utilities.logging import LogOwner
from .utilities import add_kwargs_to_configs, KCCommitOffsetDictKey, KCCommitOffset
from .config_file_parser import KafkaConfigFileParser
from .openmsistream_kafka_crypto import OpenMSIStreamKafkaCrypto
from .serialization import CompoundDeserializer

class OpenMSIStreamConsumer(LogOwner) :
    """
    Convenience class for working with a Consumer of some type
    """

    def __init__(self,consumer_type,configs,kafkacrypto=None,**kwargs) :
        """
        consumer_type = the type of Consumer underlying this object
        configs = a dictionary of configurations to pass to the consumer to create it
        """
        super().__init__(**kwargs)
        if consumer_type==KafkaConsumer :
            if kafkacrypto is None :
                self.logger.error('ERROR: creating a KafkaConsumer requires holding onto its KafkaCrypto objects!')
            self.__kafkacrypto = kafkacrypto
            self.__consumer = consumer_type(**configs)
            self.__messages = []
        elif consumer_type==DeserializingConsumer :
            self.__consumer = consumer_type(configs)
        else :
            errmsg=f'ERROR: Unrecognized consumer type {consumer_type} for OpenMSIStreamConsumer!'
            self.logger.error(errmsg,ValueError)
        self.configs = configs

    @staticmethod
    def get_consumer_args_kwargs(config_file_path,logger=None,**kwargs) :
        """
        Return the arguments and keyword arguments that should be used to create 
        a OpenMSIStreamConsumer based on the configs

        config_file_path = path to the config file to use in defining this consumer

        any keyword arguments will be added to the final consumer configs (with underscores replaced with dots)

        Used to quickly instantiate more than one identical OpenMSIStreamConsumer for a ConsumerGroup
        """
        parser = KafkaConfigFileParser(config_file_path,logger=logger)
        ret_kwargs = {}
        #get the broker and consumer configurations
        all_configs = {**parser.broker_configs,**parser.consumer_configs}
        all_configs = add_kwargs_to_configs(all_configs,logger,**kwargs)
        #all_configs['debug']='broker,topic,msg'
        #if the group.id has been set as "create_new" generate a new group ID
        if 'group.id' in all_configs.keys() and all_configs['group.id'].lower()=='create_new' :
            all_configs['group.id']=str(uuid.uuid1())
        #if there are configs for KafkaCrypto, use a KafkaConsumer
        if parser.kc_config_file_str is not None :
            if logger is not None :
                logger.debug(f'Consumed messages will be decrypted using configs at {parser.kc_config_file_str}')
            kc = OpenMSIStreamKafkaCrypto(parser.broker_configs,parser.kc_config_file_str)
            if 'key.deserializer' in all_configs.keys() :
                keydes = CompoundDeserializer(kc.key_deserializer,all_configs.pop('key.deserializer'))
            else :
                keydes = kc.key_deserializer
            if 'value.deserializer' in all_configs.keys() :
                valdes = CompoundDeserializer(kc.value_deserializer,all_configs.pop('value.deserializer'))
            else :
                valdes = kc.value_deserializer
            all_configs['key_deserializer']=keydes
            all_configs['value_deserializer']=valdes
            ret_args = [KafkaConsumer,all_configs]
            ret_kwargs['kafkacrypto']=kc
        #otherwise use a DeserializingConsumer
        else :
            ret_args = [DeserializingConsumer,all_configs]
        ret_kwargs['logger'] = logger
        return ret_args, ret_kwargs

    @classmethod
    def from_file(cls,*args,**kwargs) :
        args_to_use, kwargs_to_use = OpenMSIStreamConsumer.get_consumer_args_kwargs(*args,**kwargs)
        return cls(*args_to_use,**kwargs_to_use)

    def get_next_message(self,*poll_args,**poll_kwargs) :
        """
        Call "poll" for this consumer and return any successfully consumed message's value
        otherwise log a warning if there's an error while calling poll

        For the case of encrypted messages, this may return a Message object with 
        KafkaCryptoMessages for its key and/or value if the message fails 
        to be decrypted within a certain amount of time
        """ 
        #There's one version for the result of a KafkaConsumer.poll() call
        if isinstance(self.__consumer,KafkaConsumer) :
            #check if there are any messages still waiting to be processed from a recent KafkaCrypto poll call
            if isinstance(self.__consumer,KafkaConsumer) and len(self.__messages)>0 :
                consumed_msg = self.__messages.pop(0)
                return consumed_msg
            msg_dict = self.__consumer.poll(*poll_args,**poll_kwargs)
            if msg_dict=={} :
                return
            for pk in msg_dict.keys() :
                for m in msg_dict[pk] :
                    self.__messages.append(m)
            return self.get_next_message(*poll_args,**poll_kwargs)
        #And another version for a regular Consumer
        else :
            consumed_msg = None
            try :
                consumed_msg = self.__consumer.poll(*poll_args,**poll_kwargs)
            except Exception as e :
                warnmsg = 'WARNING: encountered an error in a call to consumer.poll() '
                warnmsg+= f'and this message will be skipped. Exception: {e}'
                self.logger.warning(warnmsg)
                #raise e
                return None
            if consumed_msg is not None and consumed_msg!={} :
                if consumed_msg.error() is not None or consumed_msg.value() is None :
                    warnmsg = f'WARNING: unexpected consumed message, consumed_msg = {consumed_msg}'
                    warnmsg+= f', consumed_msg.error() = {consumed_msg.error()}, '
                    warnmsg+= f'consumed_msg.value() = {consumed_msg.value()}'
                    self.logger.warning(warnmsg)
                return consumed_msg
            else :
                return None

    def subscribe(self,*args,**kwargs) :
        self.__consumer.subscribe(*args,**kwargs)

    def commit(self,message=None,offsets=None,asynchronous=True) :
        if (message is not None) and (not isinstance(message,Message)) :
            try :
                offset_dict = {KCCommitOffsetDictKey(message.topic,message.partition):KCCommitOffset(message.offset)}
                if asynchronous :
                    return self.__consumer.commit_async(offsets=offset_dict)
                else :
                    return self.__consumer.commit(offsets=offset_dict)
            except :
                warnmsg = 'WARNING: failed to commit an offset for an encrypted message. '
                warnmsg = 'Duplicates may result if the Consumer is restarted.'
                self.logger.warning(warnmsg)
                return None
        if message is None :
            return self.__consumer.commit(offsets=offsets,asynchronous=asynchronous)
        elif offsets is None :
            return self.__consumer.commit(message=message,asynchronous=asynchronous)
        else :
            errmsg = 'ERROR: "message" and "offset" arguments are exclusive for Consumer.commit. Nothing commited.'
            self.logger.error(errmsg)
    
    def close(self,*args,**kwargs) :
        self.__consumer.close(*args,**kwargs)
        try :
            if self.__kafkacrypto :
                self.__kafkacrypto.close()
        except :
            pass
        finally :
            self.__kafkacrypto = None



#imports
from collections import namedtuple

def add_kwargs_to_configs(configs,logger,**kwargs) :
    """
    Add any kwargs with underscores replaced with dots to a given config dictionary
    """
    new_configs = configs.copy()
    for argname,arg in kwargs.items() :
        config_name = argname.replace('_','.')
        if config_name in new_configs.keys() and new_configs[config_name]!=arg :
            if logger is not None :
                warnmsg = f'WARNING: a new value for the "{config_name}" config has been supplied '
                warnmsg+=  'by a keyword argument that will overwrite a previously-set value. '
                warnmsg+= f'The value will be set to {arg} instead of {new_configs[config_name]}.'
                logger.warning(warnmsg)
        new_configs[config_name]=arg
    return new_configs

#a callback function to use for testing whether a message has been successfully produced to the topic
def default_producer_callback(err,msg,logger=None,**other_kwargs) :
    if err is not None: #raise an error if the message wasn't sent successfully
        if err.fatal() :
            logmsg =f'ERROR: fatally failed to deliver message with kwargs "{other_kwargs}". '
            logmsg+=f'MESSAGE DROPPED. Error reason: {err.str()}'
            if logger is not None :
                logger.error(logmsg)
            else :
                raise RuntimeError(logmsg)
        elif not err.retriable() :
            logmsg =f'ERROR: Failed to deliver message with kwargs "{other_kwargs}" and cannot retry. '
            logmsg+= f'MESSAGE DROPPED. Error reason: {err.str()}'
            if logger is not None :
                logger.error(logmsg)
            else :
                raise RuntimeError(logmsg)

def make_callback(func,*args,**kwargs) :
    return lambda err,msg : func(err,msg,*args,**kwargs)


KCCommitOffsetDictKey = namedtuple('KCCommitOffsetDictKey',['topic','partition'])
KCCommitOffset = namedtuple('KCCommitOffset',['offset'])
